#    "Commons Clause" License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, "Sell" means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting) a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#
from typing import Callable

from PySide6.QtCore import QRunnable, Slot, Signal, QObject


class workerSignals(QObject):
    """Defines the signals available from a running worker thread."""
    finished = Signal()
    error = Signal(tuple)
    result = Signal(object)


class startThread(QRunnable):
    """A thread class to execute a given function as a runnable task."""
    __slots__ = ("fn", "args", "kwargs", "signals")

    def __init__(self, fn: Callable, *args, **kwargs) -> None:
        super().__init__()
        self.fn = fn
        self.args = args
        self.kwargs = kwargs
        self.signals = workerSignals()

    @Slot()
    def run(self) -> None:
        try:
            result = self.fn(*self.args, **self.kwargs)
            if result:
                self.signals.result.emit(result)
        except Exception as e:
            self.signals.error.emit((type(e), e.args, str(e)))
        finally:
            self.signals.finished.emit()



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting) a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#
import pathlib
import re
import sys
from pathlib import Path

from PySide6.QtCore import QObject
from PySide6.QtCore import (Signal, )
from PySide6.QtGui import (QTextCharFormat, QColor, QFont, QSyntaxHighlighter, QAction,
                           QIcon, )
from PySide6.QtWidgets import (QApplication, QMainWindow, QTextEdit, QFileDialog, QToolBar,
                               QFontDialog, QInputDialog, QMessageBox, QLabel, )

import revedaEditor.backend.dataDefinitions as ddef


class BaseHighlighter(QSyntaxHighlighter):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.highlightingRules = []
        self.commentFormat = QTextCharFormat()
        self.commentFormat.setForeground(QColor("#007F00"))

    def highlightBlock(self, text):
        for pattern, format in self.highlightingRules:
            expression = re.compile(pattern)
            for match in expression.finditer(text):
                start, end = match.span()
                self.setFormat(start, end - start, format)
        self.highlightComments(text)

    def highlightComments(self, text):
        pass


class JsonHighlighter(BaseHighlighter):
    def __init__(self, parent=None):
        super().__init__(parent)

        # String format (green)
        stringFormat = QTextCharFormat()
        stringFormat.setForeground(QColor("#009900"))
        self.highlightingRules.append((r'"(?:\\.|[^"\\])*"', stringFormat))

        # Number format (blue)
        numberFormat = QTextCharFormat()
        numberFormat.setForeground(QColor("#0000FF"))
        self.highlightingRules.append(
            (r'\b-?\d+(?:\.\d+)?(?:[eE][+-]?\d+)?\b', numberFormat))

        # Boolean format (purple)
        booleanFormat = QTextCharFormat()
        booleanFormat.setForeground(QColor("#990099"))
        self.highlightingRules.append((r'\b(true|false)\b', booleanFormat))

        # Null format (red)
        nullFormat = QTextCharFormat()
        nullFormat.setForeground(QColor("#FF0000"))
        self.highlightingRules.append((r'\bnull\b', nullFormat))

        # Object key format (brown)
        keyFormat = QTextCharFormat()
        keyFormat.setForeground(QColor("#8B4513"))
        keyFormat.setFontWeight(QFont.Bold)
        self.highlightingRules.append((r'"([^"\\]|\\.)*"\s*:', keyFormat))

        # Braces and brackets format (dark gray)
        bracesFormat = QTextCharFormat()
        bracesFormat.setForeground(QColor("#666666"))
        bracesFormat.setFontWeight(QFont.Bold)
        self.highlightingRules.append((r'[{}\[\]]', bracesFormat))


class VerilogAHighlighter(BaseHighlighter):
    def __init__(self, parent=None):
        super().__init__(parent)
        keywordFormat = QTextCharFormat()
        keywordFormat.setForeground(QColor("#0000FF"))
        keywordFormat.setFontWeight(QFont.Bold)

        keywords = ["module", "endmodule", "input", "output", "inout", "parameter",
                    "analog", "real", "electrical", "discipline", "nature"]
        self.highlightingRules.append((r'\b(' + '|'.join(keywords) + r')\b', keywordFormat))

        functionFormat = QTextCharFormat()
        functionFormat.setForeground(QColor("#00FF00"))
        functionFormat.setFontWeight(QFont.Bold)

        functions = ["cos", "sin", "ln", "log", "min", "pow", "sinh", "sqrt", "tan", "tanh",
                     "exp", "cosh", "ddt", "ddx", "idt", "laplace_nd", "laplace_zd"]
        self.highlightingRules.append(
            (r'\b(' + '|'.join(functions) + r')\b', functionFormat))

    def highlightComments(self, text):
        self.highlightSingleLineComments(text, '//')
        self.highlightMultiLineComments(text, r'/\*', r'\*/')

    def highlightSingleLineComments(self, text, commentStart):
        expression = re.compile(commentStart + '.*$')
        for match in expression.finditer(text):
            start, end = match.span()
            self.setFormat(start, end - start, self.commentFormat)

    def highlightMultiLineComments(self, text, commentStart, commentEnd):
        startIndex = 0
        if self.previousBlockState() != 1:
            startIndex = text.find(commentStart)

        while startIndex >= 0:
            endIndex = text.find(commentEnd, startIndex)
            if endIndex == -1:
                self.setCurrentBlockState(1)
                commentLength = len(text) - startIndex
            else:
                commentLength = endIndex - startIndex + len(commentEnd)

            self.setFormat(startIndex, commentLength, self.commentFormat)
            startIndex = text.find(commentStart, startIndex + commentLength)


class XyceHighlighter(BaseHighlighter):
    def __init__(self, parent=None):
        super().__init__(parent)
        keywordFormat = QTextCharFormat()
        keywordFormat.setForeground(QColor("#0000FF"))
        keywordFormat.setFontWeight(QFont.Bold)

        keywords = [".print", ".plot", ".include", ".subckt", ".end", ".ends", ".param",
                    ".model", ".lib", ".sweep"]
        self.highlightingRules.append((r'\b(' + '|'.join(keywords) + r')\b', keywordFormat))

        analysisFormat = QTextCharFormat()
        analysisFormat.setForeground(QColor("#00FF0F"))
        analysisFormat.setFontWeight(QFont.Bold)

        analyses = [".ac", ".dc", ".tran", ".hb", ".noise", ".op"]
        self.highlightingRules.append(
            (r'\b(' + '|'.join(analyses) + r')\b', analysisFormat))

        componentFormat = QTextCharFormat()
        componentFormat.setForeground(QColor("#FF0000"))

        components = "RCLVIQMDJBEFGHKTSWZUOPXx"
        self.highlightingRules.append((r'^[' + components + r']', componentFormat))

    def highlightComments(self, text):
        self.highlightSingleLineComments(text, r'\*')

    def highlightSingleLineComments(self, text, commentStart):
        expression = re.compile('^' + commentStart + '.*$')
        for match in expression.finditer(text):
            start, end = match.span()
            self.setFormat(start, end - start, self.commentFormat)


class textEditor(QMainWindow):
    closedSignal = Signal(QObject)

    def __init__(self, filePathObj: Path):
        super().__init__()
        self.filePathObj = filePathObj if filePathObj.exists() else pathlib.Path.cwd()
        self.textEdit = QTextEdit()
        self.setCentralWidget(self.textEdit)
        self.appMainW = QApplication.instance().appMainW

        self.createActions()
        self.createMenus()
        self.createToolbar()

        self.statusLine = self.statusBar()
        self.statusLabel = QLabel()
        self.statusLine.addPermanentWidget(self.statusLabel)
        self.textEdit.cursorPositionChanged.connect(self.updateStatus)
        self.cellViewTuple = ddef.viewNameTuple("", "", "")
        self.resize(600, 800)
        self.initEditor()

    def initEditor(self):
        if self.filePathObj.exists():
            with self.filePathObj.open("r") as file:
                text = file.read()
                self.textEdit.setPlainText(text)

    def createActions(self):
        documentOpenIcon = QIcon(":/icons/document-task.png")
        self.openAction = QAction(documentOpenIcon, "&Open", self)
        self.openAction.setShortcut("Ctrl+O")
        self.openAction.setStatusTip("Open File")
        self.openAction.triggered.connect(self.openFile)

        reloadIcon = QIcon(":/icons/arrow-circle.png")
        self.reloadAction = QAction(reloadIcon, "&Reload", self)
        self.reloadAction.setShortcut("F5")
        self.reloadAction.setStatusTip("Reload File")
        self.reloadAction.triggered.connect(self.reloadFile)

        saveIcon = QIcon(":/icons/document.png")
        self.saveAction = QAction(saveIcon, "&Save", self)
        self.saveAction.setShortcut("Ctrl+S")
        self.saveAction.setStatusTip("Save File")
        self.saveAction.triggered.connect(self.saveFile)

        saveAsIcon = QIcon(":/icons/document--plus.png")
        self.saveAsAction = QAction(saveAsIcon, "Save &As", self)
        self.saveAsAction.setStatusTip("Save File As...")
        self.saveAsAction.triggered.connect(self.saveAsFile)

        quitIcon = QIcon(":/icons/external.png")
        self.quitAction = QAction(quitIcon, "&Quit", self)
        self.quitAction.setShortcut("Ctrl+Q")
        self.quitAction.setStatusTip("Quit")
        self.quitAction.triggered.connect(self.close)

        fontIcon = QIcon(":icons/ui-label.png")
        self.fontAction = QAction(fontIcon, "&Font", self)
        self.fontAction.setShortcut("Ctrl+F")
        self.fontAction.setStatusTip("Select Font")
        self.fontAction.triggered.connect(self.changeFont)

        cutIcon = QIcon(":/icons/cutter.png")
        self.cutAction = QAction(cutIcon, "&Cut", self)
        self.cutAction.setShortcut("Ctrl+X")
        self.cutAction.setStatusTip("Cut Text")
        self.cutAction.triggered.connect(self.textEdit.cut)

        copyIcon = QIcon(":/icons/document-copy.png")
        self.copyAction = QAction(copyIcon, "&Copy", self)
        self.copyAction.setShortcut("Ctrl+C")
        self.copyAction.setStatusTip("Copy Text")
        self.copyAction.triggered.connect(self.textEdit.copy)

        pasteIcon = QIcon(":/icons/clipboard-paste.png")
        self.pasteAction = QAction(pasteIcon, "&Paste", self)
        self.pasteAction.setShortcut("Ctrl+V")
        self.pasteAction.setStatusTip("Paste Text")
        self.pasteAction.triggered.connect(self.textEdit.paste)

        findIcon = QIcon(":/icons/clipboard--pencil.png")
        self.findAction = QAction(findIcon, "&Find", self)
        self.findAction.setShortcut("Ctrl+F")
        self.findAction.setStatusTip("Find Text")
        self.findAction.triggered.connect(self.findText)

        replaceIcon = QIcon(":/icons/clipboard-search-result.png")
        self.replaceAction = QAction(replaceIcon, "&Replace", self)
        self.replaceAction.setShortcut("Ctrl+H")
        self.replaceAction.setStatusTip("Replace Text")
        self.replaceAction.triggered.connect(self.replaceText)

    def createMenus(self):
        menubar = self.menuBar()
        menubar.setNativeMenuBar(False)
        fileMenu = menubar.addMenu("&File")
        fileMenu.addAction(self.openAction)
        fileMenu.addAction(self.reloadAction)
        fileMenu.addAction(self.saveAction)
        fileMenu.addAction(self.saveAsAction)
        fileMenu.addAction(self.quitAction)

        format_menu = menubar.addMenu("&Format")
        format_menu.addAction(self.fontAction)

        editMenu = menubar.addMenu("&Edit")
        editMenu.addAction(self.cutAction)
        editMenu.addAction(self.copyAction)
        editMenu.addAction(self.pasteAction)
        editMenu.addAction(self.findAction)
        editMenu.addAction(self.replaceAction)

    def createToolbar(self):
        toolbar = QToolBar("Toolbar")
        self.addToolBar(toolbar)

        toolbar.addAction(self.openAction)
        toolbar.addAction(self.reloadAction)
        toolbar.addAction(self.saveAction)
        toolbar.addAction(self.saveAsAction)
        toolbar.addAction(self.quitAction)
        toolbar.addSeparator()
        toolbar.addAction(self.fontAction)
        toolbar.addSeparator()
        toolbar.addAction(self.cutAction)
        toolbar.addAction(self.copyAction)
        toolbar.addAction(self.pasteAction)
        toolbar.addSeparator()
        toolbar.addAction(self.findAction)
        toolbar.addAction(self.replaceAction)

    def openFile(self):
        (fileName, _) = QFileDialog.getOpenFileName(self, "Open File",
                                                    str(self.filePathObj),
                                                    "JSON Files (*.json);;All Files (*)")
        if fileName:
            self.filePathObj = Path(fileName)
            if self.filePathObj.exists():
                with self.filePathObj.open("r") as file:
                    text = file.read()
                    self.textEdit.setPlainText(text)

    def saveFile(self):
        if self.filePathObj.exists():
            with self.filePathObj.open("w") as file:
                text = self.textEdit.toPlainText()
                file.write(text)
        else:
            self.saveAsFile()

    def saveAsFile(self):
        (fileName, _) = QFileDialog.getSaveFileName(self, "Save File",
                                                    str(self.filePathObj),
                                                    "JSON Files (*.json);;All Files (*)")
        if fileName:
            self.filePathObj = Path(fileName)
            if self.filePathObj.exists():
                with self.filePathObj.open("w") as file:
                    text = self.textEdit.toPlainText()
                    file.write(text)

    def reloadFile(self):
        if self.filePathObj.exists():
            cursor_position = self.textEdit.textCursor().position()
            with self.filePathObj.open("r") as file:
                text = file.read()
                self.textEdit.setPlainText(text)
            # Try to restore cursor position if possible
            cursor = self.textEdit.textCursor()
            cursor.setPosition(min(cursor_position, len(self.textEdit.toPlainText())))
            self.textEdit.setTextCursor(cursor)
        else:
            QMessageBox.information(self, "Reload", "No file is currently open.")

    def changeFont(self):
        ok, font = QFontDialog.getFont(self.textEdit.font(), self)
        if ok:
            self.textEdit.setFont(font)

    def findText(self):
        text, ok = QInputDialog.getText(self, "Find", "Enter text to find:")
        if ok:
            document = self.textEdit.document()
            cursor = document.find(text)
            if cursor.isNull():
                QMessageBox.information(self, "Find", "Text not found.")
            else:
                self.textEdit.setTextCursor(cursor)

    def replaceText(self):
        findText, ok1 = QInputDialog.getText(self, "Replace", "Enter text to find:")
        if ok1:
            replace_text, ok2 = QInputDialog.getText(self, "Replace",
                                                     "Enter replacement text:")
            if ok2:
                document = self.textEdit.document()
                cursor = document.find(findText)
                if cursor.isNull():
                    QMessageBox.information(self, "Replace", "Text not found.")
                else:
                    cursor.insertText(replace_text)

    def updateStatus(self):
        cursor = self.textEdit.textCursor()
        line = cursor.blockNumber() + 1
        column = cursor.columnNumber() + 1
        self.statusLabel.setText(f"Line: {line}, Column: {column}")

    def closeEvent(self, event):
        self.closedSignal.emit(self)
        self.appMainW.openViews.pop(self.cellViewTuple, None)
        super().closeEvent(event)


class jsonEditor(textEditor):
    def __init__(self, filePathObj: Path):
        super().__init__(filePathObj)
        self.initEditor()
        self.setWindowTitle("JSON Editor")

    def initEditor(self):
        self.highlighter = JsonHighlighter(self.textEdit.document())
        super().initEditor()


class verilogaEditor(textEditor):
    def __init__(self, filePathObj):
        super().__init__(filePathObj)
        self.initEditor()
        self.setWindowTitle("Verilog-A Editor")

    def initEditor(self):
        self.highlighter = VerilogAHighlighter(self.textEdit.document())
        super().initEditor()

    def openFile(self):
        (fileName, _) = QFileDialog.getOpenFileName(self, "Open File", "",
                                                    "Verilog-A Files (*.va);;All Files (*)")
        if fileName:
            self.filePathObj = Path(fileName)
            if self.filePathObj.exists():
                with self.filePathObj.open("rw") as file:
                    text = file.read()
                    self.textEdit.setPlainText(text)

    def saveAsFile(self):
        (fileName, _) = QFileDialog.getSaveFileName(self, "Save File",
                                                    str(self.filePathObj),
                                                    "Verilog-A Files (*.va);;All Files (*)")
        if fileName:
            self.filePathObj = Path(fileName)
            if self.filePathObj.exists():
                with self.filePathObj.open("w") as file:
                    text = self.textEdit.toPlainText()
                    file.write(text)


class spiceEditor(textEditor):
    def __init__(self, filePathObj: Path):
        super().__init__(filePathObj)
        self.initEditor()
        self.setWindowTitle("Xyce/SPICE Editor")

    def initEditor(self):
        self.highlighter = XyceHighlighter(self.textEdit.document())
        super().initEditor()

    def openFile(self):
        (fileName, _) = QFileDialog.getOpenFileName(self, "Open File", "",
                                                    "Xyce Files (*.sp);;All Files (*)")
        if fileName:
            self.filePathObj = Path(fileName)
            if self.filePathObj.exists():
                with self.filePathObj.open("r") as file:
                    text = file.read()
                    self.textEdit.setPlainText(text)

    def saveAsFile(self):

        (fileName, _) = QFileDialog.getSaveFileName(self, "Save File",
                                                    str(self.filePathObj),
                                                    "Xyce Files (*.sp);;All Files (*)")
        if fileName:
            self.filePathObj = Path(fileName)
            if self.filePathObj:
                with self.filePathObj.open("w") as file:
                    text = self.textEdit.toPlainText()
                    file.write(text)


def main():
    app = QApplication(sys.argv)
    editor = verilogaEditor(None, '')
    editor.show()
    sys.exit(app.exec())


if __name__ == "__main__":
    main()



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting) a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#
from PySide6.QtCore import Qt
from PySide6.QtGui import QImage, QPainter, QPen, QAction, QIcon
from PySide6.QtWidgets import (
    QGraphicsScene,
    QGraphicsView,
    QMainWindow,
    QToolBar,
    QFileDialog,
    QComboBox,
    QLabel,
    QMessageBox,
)


class stippleView(QGraphicsView):
    def __init__(self, scene: QGraphicsScene, parent=None, size: int = 32):
        super().__init__(scene, parent)
        self._size = size
        self._gridStep = 20
        self.setRenderHint(QPainter.Antialiasing)
        # self.setDragMode(QGraphicsView.ScrollHandDrag)
        self.setCursor(Qt.ArrowCursor)

        self.setSceneRect(
            0, 0, self._size * self._gridStep, self._size * self._gridStep
        )
        self.drawGrid()

        self.gridSquares = [[None] * self._size for _ in range(self._size)]

    def drawGrid(self):
        pen = QPen(Qt.lightGray)
        for x in range(0, self._size * self._gridStep + 1, self._gridStep):
            self.scene().addLine(x, 0, x, self._size * self._gridStep, pen)
        for y in range(0, self._size * self._gridStep + 1, self._gridStep):
            self.scene().addLine(0, y, self._size * self._gridStep, y, pen)

    def drawDot(self, row, col):
        try:
            square = self.scene().addRect(
                col * self._gridStep,
                row * self._gridStep,
                self._gridStep,
                self._gridStep,
                Qt.NoPen,
                Qt.black,
            )
            self.gridSquares[row][col] = square
        except IndexError:
            msgBox = QMessageBox()
            msgBox.setIcon(QMessageBox.Critical)
            msgBox.setWindowTitle("Error")
            msgBox.setText("Please select a square inside the grid.")
            msgBox.setStandardButtons(QMessageBox.Ok)
            msgBox.exec()
            self.scene().removeItem(square)

    def clearDot(self, row, col):
        square = self.gridSquares[row][col]
        if square is not None:
            self.scene().removeItem(square)
            self.gridSquares[row][col] = None

    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton:
            pos = self.mapToScene(event.pos())
            col = int(pos.x() // self._gridStep)
            row = int(pos.y() // self._gridStep)
            try:
                if self.gridSquares[row][col] is not None:
                    self.clearDot(row, col)
                else:
                    self.drawDot(row, col)
            except IndexError:
                pass
        else:
            super().mousePressEvent(event)

    @property
    def size(self):
        return self._size

    @size.setter
    def size(self, size):
        self._size = size

    @property
    def gridStep(self):
        return self._gridStep


class stippleScene(QGraphicsScene):
    def __init__(self, parent):
        super().__init__(parent)


class stippleEditor(QMainWindow):
    def __init__(self, parent):
        super().__init__(parent)
        self.setWindowTitle("Stipple Editor")
        self.scene = stippleScene(self)
        self.view = stippleView(self.scene, self, 32)
        self.setCentralWidget(self.view)
        self.sizeCB = QComboBox(self)
        self._createActions()
        self._createMenuBar()
        self._createToolBar()

    def _createActions(self):
        clearIcon = QIcon(":/icons/node-delete.png")
        self.clearAction = QAction(clearIcon, "Clear", self)
        self.clearAction.triggered.connect(self.clearPattern)
        self.clearAction.setToolTip("Clear Pattern")

        exportImageIcon = QIcon(":/icons/image-export.png")
        self.exportImageAction = QAction(exportImageIcon, "Export as Image", self)
        self.exportImageAction.triggered.connect(self.exportPatternAsImage)
        self.exportImageAction.setToolTip("Export Pattern as Image")

        exportTextIcon = QIcon(":/icons/script-text.png")
        self.exportTextAction = QAction(exportTextIcon, "Save Pattern", self)
        self.exportTextAction.triggered.connect(self.exportPatternAsText)
        self.exportTextAction.setToolTip("Save Pattern as Text File")

        loadTextIcon = QIcon(":/icons/property-blue.png")
        self.loadTextAction = QAction(loadTextIcon, "Load...", self)
        self.loadTextAction.triggered.connect(self.loadPatternFromText)
        self.loadTextAction.setToolTip("Load Pattern from File")

        exitIcon = QIcon(":/icons/external.png")
        self.exitAction = QAction(exitIcon, "Close Window", self)
        self.exitAction.setShortcut("Ctrl+Q")

    def _createMenuBar(self):
        self.editorMenuBar = self.menuBar()
        self.editorMenuBar.setNativeMenuBar(False)
        # Returns QMenu object.
        self.menuFile = self.editorMenuBar.addMenu("&File")
        self.menuFile.addAction(self.loadTextAction)
        self.menuFile.addAction(self.exportTextAction)
        self.menuFile.addAction(self.exportImageAction)
        self.menuEdit = self.editorMenuBar.addMenu("&Edit")
        self.menuEdit.addAction(self.clearAction)

    def _createToolBar(self):
        toolbar = QToolBar(self)
        toolbar.addAction(self.clearAction)
        toolbar.addAction(self.exportImageAction)
        toolbar.addAction(self.exportTextAction)
        toolbar.addAction(self.loadTextAction)

        self.sizeCB.addItems(["8x8", "16x16", "32x32"])
        self.sizeCB.setCurrentIndex(2)
        self.sizeCB.currentIndexChanged.connect(self.matrixSizeChanged)
        toolbar.addWidget(QLabel("Stipple Matrix Size: "))
        toolbar.addWidget(self.sizeCB)
        self.addToolBar(toolbar)

    def clearPattern(self):
        self.view.scene().clear()
        self.view.gridSquares = [[None] * self.view.size for _ in range(self.view.size)]
        self.view.drawGrid()

    def exportPatternAsImage(self):
        file_dialog = QFileDialog(self)
        file_dialog.setAcceptMode(QFileDialog.AcceptSave)
        file_dialog.setDefaultSuffix("png")
        file_dialog.setNameFilter("PNG Image (*.png)")
        if file_dialog.exec():
            file_path = file_dialog.selectedFiles()[0]
            self.imageExportToFile(file_path)

    def imageExportToFile(self, file_path):
        image = QImage(
            self.view.size * self.view.gridStep,
            self.view.size * self.view.gridStep,
            QImage.Format_RGB32,
        )
        image.fill(Qt.white)
        painter = QPainter(image)
        painter.setRenderHint(QPainter.Antialiasing)
        for row in self.view.gridSquares:
            for square in row:
                if square is not None:
                    painter.fillRect(square.rect(), Qt.black)
        painter.end()
        image.save(file_path)

    def exportPatternAsText(self):
        file_dialog = QFileDialog(self)
        file_dialog.setAcceptMode(QFileDialog.AcceptSave)
        file_dialog.setDefaultSuffix("txt")
        file_dialog.setNameFilter("Text File (*.txt)")
        if file_dialog.exec():
            file_path = file_dialog.selectedFiles()[0]

            pattern = []
            for row in range(self.view.size):
                row_data = []
                for col in range(self.view.size):
                    square = self.view.gridSquares[row][col]
                    if square is not None:
                        row_data.append("1")
                    else:
                        row_data.append("0")
                pattern.append(" ".join(row_data))

            with open(file_path, "w") as file:
                file.write("\n".join(pattern))

    def loadPatternFromText(self):
        file_dialog = QFileDialog(self)
        file_dialog.setAcceptMode(QFileDialog.AcceptOpen)
        file_dialog.setDefaultSuffix("txt")
        file_dialog.setNameFilter("Text File (*.txt)")
        if file_dialog.exec():
            file_path = file_dialog.selectedFiles()[0]

            self.clearPattern()

            self.loadPatternFromFile(file_path)

    def loadPatternFromFile(self, file_path):
        with open(file_path, "r") as file:
            lines = file.readlines()
        index = 2
        match len(lines):
            case 8:
                index = 0
            case 16:
                index = 1
            case 32:
                index = 2
        self.sizeCB.setCurrentIndex(index)
        self.matrixSizeChanged(index)
        for row, line in enumerate(lines):
            row_data = line.strip().split()
            for col, data in enumerate(row_data):
                if data == "1":
                    self.view.drawDot(row, col)

    def matrixSizeChanged(self, index: int):
        self.view.scene().clear()
        match index:
            case 0:
                self.view.size = 8
            case 1:
                self.view.size = 16
            case 2:
                self.view.size = 32
        self.view.gridSquares = [[None] * self.view.size for _ in range(self.view.size)]
        self.view.drawGrid()



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting) a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#

from PySide6.QtCore import Qt, QPoint
from PySide6.QtWidgets import QMainWindow, QDialog

import revedaEditor.backend.dataDefinitions as ddef
import revedaEditor.backend.hdlBackEnd as hdl
# import revedaEditor.backend.libBackEnd as scb  # import the backend
import revedaEditor.common.shapes as shp
from revedaEditor.backend import dataDefinitions as ddef, hdlBackEnd as hdl, \
    libBackEnd as scb
from revedaEditor.fileio import symbolEncoder as se
from revedaEditor.gui import libraryBrowser as libw, fileDialogues as fd, \
    symbolEditor as syed, \
    propertyDialogues as pdlg


def drawBaseSymbol(symbolScene, dlg):
    leftPinNames = list(
        filter(
            None,
            [pinName.strip() for pinName in dlg.leftPinsEdit.text().split(",")],
        )
    )
    rightPinNames = list(
        filter(
            None,
            [pinName.strip() for pinName in dlg.rightPinsEdit.text().split(",")],
        )
    )
    topPinNames = list(
        filter(
            None,
            [pinName.strip() for pinName in dlg.topPinsEdit.text().split(",")],
        )
    )
    bottomPinNames = list(
        filter(
            None,
            [pinName.strip() for pinName in dlg.bottomPinsEdit.text().split(",")],
        )
    )
    stubLength = (
        int(float(dlg.stubLengthEdit.text().strip()))
        if dlg.stubLengthEdit.text()
        else 60
    )
    pinDistance = (
        int(float(dlg.pinDistanceEdit.text().strip()))
        if dlg.pinDistanceEdit.text()
        else 80
    )
    rectXDim = (max(len(topPinNames), len(bottomPinNames)) + 1) * pinDistance
    rectYDim = (max(len(leftPinNames), len(rightPinNames)) + 1) * pinDistance

    symbolScene.rectDraw(
        QPoint(0, 0),
        QPoint(rectXDim, rectYDim),
    )
    symbolScene.labelDraw(
        QPoint(int(0.25 * rectXDim), int(0.4 * rectYDim)),
        "[@cellName]",
        "NLPLabel",
        "12",
        "Center",
        "R0",
        "Instance",
    )
    symbolScene.labelDraw(
        QPoint(int(rectXDim), int(-0.1 * rectYDim)),
        "[@instName]",
        "NLPLabel",
        "12",
        "Center",
        "R0",
        "Instance",
    )

    leftPinLocs = [
        QPoint(-stubLength, (i + 1) * pinDistance) for i in range(len(leftPinNames))
    ]
    rightPinLocs = [
        QPoint(rectXDim + stubLength, (i + 1) * pinDistance)
        for i in range(len(rightPinNames))
    ]
    bottomPinLocs = [
        QPoint((i + 1) * pinDistance, rectYDim + stubLength)
        for i in range(len(bottomPinNames))
    ]
    topPinLocs = [
        QPoint((i + 1) * pinDistance, -stubLength) for i in range(len(topPinNames))
    ]
    for i, pinName in enumerate(leftPinNames):
        symbolScene.lineDraw(
            leftPinLocs[i],
            leftPinLocs[i] + QPoint(stubLength, 0),
        )
        symbolScene.addItem(
            shp.symbolPin(
                leftPinLocs[i],
                pinName,
                shp.symbolPin.pinDirs[0],
                shp.symbolPin.pinTypes[0],
            )
        )
    for i, pinName in enumerate(rightPinNames):
        symbolScene.lineDraw(
            rightPinLocs[i],
            rightPinLocs[i] + QPoint(-stubLength, 0),
        )
        symbolScene.addItem(
            shp.symbolPin(
                rightPinLocs[i],
                pinName,
                shp.symbolPin.pinDirs[1],
                shp.symbolPin.pinTypes[0],
            )
        )
    for i, pinName in enumerate(topPinNames):
        symbolScene.lineDraw(
            topPinLocs[i],
            topPinLocs[i] + QPoint(0, stubLength),
        )
        symbolScene.addItem(
            shp.symbolPin(
                topPinLocs[i],
                pinName,
                shp.symbolPin.pinDirs[2],
                shp.symbolPin.pinTypes[2],
            )
        )
    for i, pinName in enumerate(bottomPinNames):
        symbolScene.lineDraw(
            bottomPinLocs[i],
            bottomPinLocs[i] + QPoint(0, -stubLength),
        )
        symbolScene.addItem(
            shp.symbolPin(
                bottomPinLocs[i],
                pinName,
                shp.symbolPin.pinDirs[2],
                shp.symbolPin.pinTypes[1],
            )
        )

    return rectXDim, rectYDim


def createVaSymbol(
        parent: QMainWindow,
        vaItemTuple: ddef.viewItemTuple,
        libraryDict: dict,
        libraryBrowser: libw.libraryBrowser,
        importedVaObj: hdl.verilogaC,
) -> None:
    """
    Creates a symbol for a given view item in the library.

    Args:
        parent (QMainWindow): The parentW window.
        vaItemTuple (ddef.viewItemTuple): The view item tuple.
        libraryDict (dict): The library dictionary.
        libraryBrowser (edw.libraryBrowser): The library browser.
        importedVaObj (hdl.verilogaC): The imported Veriloga object.

    Returns:
        None

    Raises:
        None
    """
    # symbolNameDlg = fd.createCellViewDialog(
    #     parent, libraryBrowser.libraryModel, vaItemTuple.cellItem
    # )
    # symbolNameDlg.viewComboBox.setCurrentText("symbol")
    # symbolNameDlg.nameEdit.setText("symbol")
    symbolNameDlg = fd.newCellViewDialog(
        parent, libraryBrowser.designView.libraryModel
    )
    symbolNameDlg.libNamesCB.setCurrentText(vaItemTuple.libraryItem.libraryName)
    symbolNameDlg.cellCB.setCurrentText(vaItemTuple.cellItem.cellName)
    symbolNameDlg.viewType.addItems(["symbol"])
    symbolNameDlg.viewName.setText("symbol")
    if symbolNameDlg.exec() == QDialog.DialogCode.Accepted:
        symbolViewName = symbolNameDlg.viewName.text().strip()
        symbolViewItem = scb.createCellView(
            parent, symbolViewName, vaItemTuple.cellItem
        )
        importedVaObjPath = vaItemTuple.cellItem.data(
            Qt.ItemDataRole.UserRole + 2).joinpath(
            importedVaObj.pathObj.name
        )
        symbolWindow = syed.symbolEditor(
            symbolViewItem,
            libraryDict,
            libraryBrowser.libBrowserCont.designView,
        )
        symbolScene = symbolWindow.centralW.scene
        dlg = pdlg.symbolCreateDialog(parent)

        dlg.leftPinsEdit.setText(",".join(importedVaObj.inPins))
        dlg.rightPinsEdit.setText(",".join(importedVaObj.outPins))
        dlg.topPinsEdit.setText(",".join(importedVaObj.inoutPins))

        if dlg.exec() == QDialog.DialogCode.Accepted:
            rectXDim, rectYDim = drawBaseSymbol(symbolScene, dlg)
            vaModuleLabel = symbolScene.labelDraw(
                QPoint(int(0.25 * rectXDim), int(0.8 * rectYDim)),
                f"[@vaModule:vaModule=%:vaModule={importedVaObj.vaModule}]",
                "NLPLabel",
                "12",
                "Center",
                "R0",
                "Instance",
            )
            vaModuleLabel.labelVisible = True

            instParamNum = len(importedVaObj.instanceParams)
            if instParamNum > 0:
                for index, (key, value) in enumerate(
                        importedVaObj.instanceParams.items()
                ):
                    symbolScene.labelDraw(
                        QPoint(
                            int(rectXDim),
                            int(index * 0.2 * rectYDim / instParamNum),
                        ),
                        f"[@{key}:{key}=%:{key}={value}]",
                        "NLPLabel",
                        "12",
                        "Center",
                        "R0",
                        "Instance",
                    )

            symbolScene.attributeList = list()  # empty attribute list
            if importedVaObj.modelParams:
                for key, value in importedVaObj.modelParams.items():
                    symbolScene.attributeList.append(se.symbolAttribute(key, value))
                symbolScene.attributeList.append(
                    se.symbolAttribute(
                        "SpiceNetlistLine", importedVaObj.netlistLine
                    )
                )

            modelParamsString = ", ".join(
                f"{key} = {value}" for key, value in importedVaObj.modelParams.items()
            )
            symbolScene.attributeList.append(
                se.symbolAttribute(
                    "vaModelLine",
                    f".MODEL {importedVaObj.vaModule}Model {importedVaObj.vaModule} {modelParamsString}",
                )
            )
            symbolScene.attributeList.append(
                se.symbolAttribute("vaHDLLine", f"*.HDL {str(importedVaObjPath)}")
            )
            symbolScene.attributeList.append(
                se.symbolAttribute("pinOrder", importedVaObj.pinOrder)
            )
            symbolWindow.show()
            symbolViewTuple = ddef.viewNameTuple(
                vaItemTuple.libraryItem.libraryName,
                vaItemTuple.cellItem.cellName,
                "symbol",
            )
            symbolWindow.libraryView.openViews[symbolViewTuple] = symbolWindow


def createSpiceSymbol(
        parent: QMainWindow,
        spiceItemTuple: ddef.viewItemTuple,
        libraryDict: dict,
        libraryBrowser: libw.libraryBrowser,
        importedSpiceObj: hdl.spiceC,
):
    symbolNameDlg = fd.newCellViewDialog(
        parent, libraryBrowser.designView.libraryModel
    )
    symbolNameDlg.libNamesCB.setCurrentText(spiceItemTuple.libraryItem.libraryName)
    symbolNameDlg.cellCB.setCurrentText(spiceItemTuple.cellItem.cellName)
    symbolNameDlg.viewType.addItems(["symbol"])
    symbolNameDlg.viewName.setText("symbol")
    if symbolNameDlg.exec() == QDialog.DialogCode.Accepted:
        symbolViewName = symbolNameDlg.viewName.text().strip()
        symbolViewItem = scb.createCellView(
            parent, symbolViewName, spiceItemTuple.cellItem
        )
        newSpiceFilePathObj = spiceItemTuple.cellItem.data(
            Qt.ItemDataRole.UserRole + 2).joinpath(
            importedSpiceObj.pathObj.name
        )
        symbolWindow = syed.symbolEditor(
            symbolViewItem,
            libraryDict,
            libraryBrowser.libBrowserCont.designView,
        )
        symbolScene = symbolWindow.centralW.scene
        dlg = pdlg.symbolCreateDialog(parent)
        dlg.leftPinsEdit.setText(", ".join(importedSpiceObj.subcktParams["pins"]))

        if dlg.exec() == QDialog.DialogCode.Accepted:
            rectXDim, rectYDim = drawBaseSymbol(symbolScene, dlg)
            symbolFileLabel = symbolScene.labelDraw(
                QPoint(int(0.25 * rectXDim), int(-0.2 * rectYDim)),
                f"[@subcktName:subcktName=%:subcktName={importedSpiceObj.subcktParams['name']}]",
                "NLPLabel",
                "12",
                "Center",
                "R0",
                "Instance",
            )
            symbolFileLabel.labelVisible = False
            instParamNum = len(importedSpiceObj.subcktParams["params"])
            for index, (key, value) in enumerate(
                    importedSpiceObj.subcktParams["params"].items()
            ):
                symbolScene.labelDraw(
                    QPoint(
                        int(rectXDim),
                        int(index * 0.2 * rectYDim / instParamNum),
                    ),
                    f"[@{key}:{key}=%:{key}={value}]",
                    "NLPLabel",
                    "12",
                    "Center",
                    "R0",
                    "Instance",
                )
            symbolScene.attributeList = list()
            symbolScene.attributeList.append(
                se.symbolAttribute("pinOrder", importedSpiceObj.pinOrder)
            )
            symbolScene.attributeList.append(
                se.symbolAttribute("incLine", importedSpiceObj.pathObj.name)
            )

            symbolScene.attributeList.append(
                se.symbolAttribute("SpiceNetlistLine",
                                   importedSpiceObj.netlistLine)
            )

            symbolWindow.show()
            symbolViewTuple = ddef.viewNameTuple(
                spiceItemTuple.libraryItem.libraryName,
                spiceItemTuple.cellItem.cellName,
                symbolViewItem.viewName,
            )
            symbolWindow.libraryView.openViews[symbolViewTuple] = symbolWindow



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting) a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#

import inspect
from pathlib import Path
from typing import List, Any

import gdstk

import revedaEditor.common.layoutShapes as lshp
from revedaEditor.backend.pdkLoader import importPDKModule

pcells = importPDKModule('pcells')


class gdsExporter:
    __slots__ = ('_cellname', '_items', '_outputFileObj', '_libraryName',
                 '_unit', '_precision', '_dbu', '_topCell', '_itemCounter', '_cellCache')

    DEFAULT_UNIT = 1e-9
    DEFAULT_PRECISION = 1e-9
    DEFAULT_DBU = 1000

    def __init__(self, cellname: str, items: List[Any], outputFileObj: Path):
        self._unit = gdsExporter.DEFAULT_UNIT
        self._precision = gdsExporter.DEFAULT_PRECISION
        self._dbu: int = gdsExporter.DEFAULT_DBU
        self._cellname = cellname
        self._items = items
        self._outputFileObj = outputFileObj
        self._libraryName = None
        self._topCell = None
        self._itemCounter = 0
        self._cellCache = {}

    def gdsExport(self):
        self._outputFileObj.parent.mkdir(parents=True, exist_ok=True)
        lib = gdstk.Library(unit=self._unit, precision=self._precision)
        self._topCell = lib.new_cell(self._cellname)  # top Cell
        for item in self._items:
            self.createCells(lib, item, self._topCell)

        lib.write_gds(self._outputFileObj)

    def gdsExportThreaded(self, threadPool):
        self._outputFileObj.parent.mkdir(parents=True, exist_ok=True)
        lib = gdstk.Library(unit=self._unit, precision=self._precision)
        self._topCell = lib.new_cell(self._cellname)
        for item in self._items:
            self.createCells(lib, item, self._topCell)

        # Run write in thread
        from revedaEditor.backend.startThread import startThread
        writer = startThread(lib.write_gds, str(self._outputFileObj))
        threadPool.start(writer)

    def oasExportThreaded(self, threadPool):
        self._outputFileObj.parent.mkdir(parents=True, exist_ok=True)
        lib = gdstk.Library(unit=self._unit, precision=self._precision)
        self._topCell = lib.new_cell(self._cellname)
        for item in self._items:
            self.createCells(lib, item, self._topCell)

        # Run write in thread
        from revedaEditor.backend.startThread import startThread
        writer = startThread(lib.write_oas, str(self._outputFileObj))
        threadPool.start(writer)

    def createCells(self, library: gdstk.Library, item: lshp.layoutShape,
                    parentCell: gdstk.Cell):
        item_type = type(item)
        if item_type == lshp.layoutInstance:
            self._processInstance(library, item, parentCell)
        elif item_type in (lshp.layoutRect, lshp.layoutPin):
            self._processRectPin(item, parentCell)
        elif item_type == lshp.layoutPath:
            self.processPath(item, parentCell)
        elif item_type == lshp.layoutLabel:
            self._processLabel(item, parentCell)
        elif item_type == lshp.layoutPolygon:
            self._processPolygon(item, parentCell)
        elif item_type == lshp.layoutViaArray:
            self._processViaArray(library, item, parentCell)
        else:
            self._process_custom_layout(library, item, parentCell)

    def _processInstance(self, library, item, parentCell):
        cellGDSName = f"{item.libraryName}_{item.cellName}_{item.viewName}_{self._itemCounter}"
        self._itemCounter += 1
        cellGDS = library.new_cell(cellGDSName)
        for shape in item.shapes:
            self.createCells(library, shape, cellGDS)
        ref = gdstk.Reference(
            cellGDS,
            (0, 0)
        )
        parentCell.add(ref)

    def _processRectPin(self, item, parentCell):
        rect = gdstk.rectangle(
            corner1=item.mapToScene(item.start).toPoint().toTuple(),
            corner2=item.mapToScene(item.end).toPoint().toTuple(),
            layer=item.layer.gdsLayer,
            datatype=item.layer.datatype,
        )
        parentCell.add(rect)

    def processPath(self, item, parentCell):
        path = gdstk.FlexPath(
            points=[item.sceneEndPoints[0].toTuple(), item.sceneEndPoints[1].toTuple()],
            width=item.width,
            ends=(item.startExtend, item.endExtend),
            simple_path=True,
            layer=item.layer.gdsLayer,
            datatype=item.layer.datatype,
        )
        parentCell.add(path)

    def _processLabel(self, item, parentCell):
        label = gdstk.Label(
            text=item.labelText,
            origin=item.mapToScene(item.start).toPoint().toTuple(),
            magnification=float(item.fontHeight * self._dbu),
            rotation=item.angle,
            layer=item.layer.gdsLayer,
            texttype=item.layer.datatype,
        )
        parentCell.add(label)

    def _processPolygon(self, item, parentCell):
        points = [item.mapToScene(point).toPoint().toTuple() for point in item.points]
        polygon = gdstk.Polygon(
            points=points,
            layer=item.layer.gdsLayer,
            datatype=item.layer.datatype,
        )
        parentCell.add(polygon)

    def _processViaArray(self, library, item, parentCell):
        via_key = (item.via.width, item.via.height, item.via.layer.name,
                   item.via.layer.purpose)
        if via_key not in self._cellCache:
            viaName = f"via_{item.via.width}_{item.via.height}_{item.via.layer.name}_{item.via.layer.purpose}"
            viaCell = library.new_cell(viaName)
            # Define single via at (0, 0) in cell-local coordinates.
            # The Reference origin carries the scene position; mixing both causes a double-offset.
            via = gdstk.rectangle(
                (0, 0),
                (item.via.width, item.via.height),
                layer=item.via.layer.gdsLayer,
                datatype=item.via.layer.datatype,
            )
            viaCell.add(via)
            self._cellCache[via_key] = viaCell
        else:
            viaCell = self._cellCache[via_key]

        viaArray = gdstk.Reference(
            cell=viaCell,
            origin=item.mapToScene(item.start).toTuple(),
            columns=item.xnum,
            rows=item.ynum,
            spacing=(item.xs + item.width, item.ys + item.height),
        )
        parentCell.add(viaArray)

    def _process_custom_layout(self, library, item, parentCell):
        if isinstance(item, pcells.baseCell):
            pcellParamDict = self.extractPcellInstanceParameters(item)
            pcellNameSuffix = "_".join(
                f"{key}_{value}".replace(".", "p") for key, value in pcellParamDict.items()
            )
            pcellName = (f"{item.libraryName}_{type(item).__name__}_"
                         f"{pcellNameSuffix}_{self._itemCounter}")
            self._itemCounter += 1
            pcellGDS = library.new_cell(pcellName)
            for shape in item.shapes:
                self.createCells(library, shape, pcellGDS)

            ref = gdstk.Reference(
                pcellGDS,
                (0, 0),
            )
            parentCell.add(ref)

    @staticmethod
    def extractPcellInstanceParameters(instance: lshp.layoutPcell) -> dict:
        initArgs = inspect.signature(instance.__class__.__init__).parameters
        argsUsed = [
            param for param in initArgs if (param != "self" and param != "snapTuple")
        ]
        argDict = {arg: getattr(instance, arg) for arg in argsUsed}
        return argDict

    @property
    def unit(self):
        return self._unit

    @unit.setter
    def unit(self, value: float):
        self._unit = value

    @property
    def precision(self):
        return self._precision

    @precision.setter
    def precision(self, value: float):
        self._precision = value

    @property
    def dbu(self):
        return self._dbu

    @dbu.setter
    def dbu(self, value: int):
        self._dbu = value



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting) a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#

import inspect
import json
import pathlib
# import time
from typing import Any, Dict, List, Union

import orjson
from PySide6.QtCore import QLineF, QPoint, QPointF, QRectF, Qt, Signal
from PySide6.QtGui import (
    QColor,
    QFont,
    QFontDatabase,
    QPen,
    QTransform,
)
from PySide6.QtWidgets import (
    QCompleter,
    QDialog,
    QGraphicsItem,
    QGraphicsLineItem,
    QGraphicsScene,
    QGraphicsSceneMouseEvent,
)

import revedaEditor.backend.dataDefinitions as ddef
import revedaEditor.backend.libraryMethods as libm
import revedaEditor.backend.libraryModelView as lmview
import revedaEditor.backend.undoStack as us
import revedaEditor.common.layoutShapes as lshp  # import layout shapes
import revedaEditor.fileio.exportGDS as gdse
import revedaEditor.fileio.layoutEncoder as layenc
import revedaEditor.fileio.loadJSON as lj
import revedaEditor.gui.editFunctions as edf
import revedaEditor.gui.fileDialogues as fd
import revedaEditor.gui.layoutDialogues as ldlg
import revedaEditor.gui.propertyDialogues as pdlg
from revedaEditor.backend.pdkLoader import importPDKModule
from revedaEditor.gui.alignItems import alignItemsDialogue, alignToLine
from revedaEditor.scenes.editorScene import editorScene

# from contextlib import contextmanager

fabproc = importPDKModule("process")
laylyr = importPDKModule("layoutLayers")
schlyr = importPDKModule("schLayers")
pcells = importPDKModule("pcells")


class layoutScene(editorScene):
    LAYOUT_SHAPES = (
        lshp.layoutRect,
        lshp.layoutPin,
        lshp.layoutLabel,
        lshp.layoutPath,
        lshp.layoutPolygon,
        lshp.layoutViaArray,
        lshp.layoutInstance,
        lshp.layoutPcell,
    )
    alignLineFinished = Signal(lshp.alignLine)

    def __init__(self, parent):
        super().__init__(parent)
        self.selectEdLayer = laylyr.pdkAllLayers[0] if laylyr else None
        # draw modes
        self.editModes = ddef.layoutModes(
            selectItem=False,
            deleteItem=False,
            moveItem=False,
            copyItem=False,
            rotateItem=False,
            changeOrigin=False,
            panView=False,
            zoomView=False,
            drawPath=False,
            drawPin=False,
            drawArc=False,
            drawPolygon=False,
            addLabel=False,
            addVia=False,
            drawRect=False,
            drawLine=False,
            drawCircle=False,
            drawRuler=False,
            stretchItem=False,
            addInstance=False,
            cutShape=False,
            alignItems=False,
        )
        self.editModes.setMode("selectItem")
        self.selectModes = ddef.layoutSelectModes(
            selectAll=True,
            selectPath=False,
            selectInstance=False,
            selectVia=False,
            selectPin=False,
            selectLabel=False,
            selectText=False,
        )
        self.messages = {
            "selectItem": "Select Item",
            "deleteItem": "Delete Item",
            "moveItem": "Move Item",
            "copyItem": "Copy Item",
            "rotateItem": "Rotate Item",
            "changeOrigin": "Change Origin",
            "panView": "Pan View at mouse Press Position",
            'zoomView': 'Draw rectangle to zoom in',
            "drawPath": "Draw Path",
            "drawPin": "Draw Pin",
            "drawArc": "Draw Arc",
            "drawPolygon": "Draw Polygon",
            "addLabel": "Add Label",
            "addVia": "Add Via",
            "drawRect": "Draw Rect",
            "drawLine": "Draw Line",
            "drawCircle": "Draw Circle",
            "drawRuler": "Draw Ruler",
            "addInstance": "Add Instance",
            "cutShape": "Cut Shape",
            "stretchItem": "Stretch Item",
            "alignItems": "Select alignment option.",
        }
        self.newInstance = None
        self.layoutInstanceTuple = None
        self._scale = fabproc.dbu if fabproc else 1000
        self.itemCounter = 0
        self.newPath = None
        self.stretchPathItem = None
        defaultPathDefTuple = fabproc.processPaths[0] if fabproc else None
        self.newPathTuple = ddef.layoutPathTuple(
            "",
            defaultPathDefTuple.layer,
            0,
            defaultPathDefTuple.minWidth,
            int(defaultPathDefTuple.minWidth / 2),
            int(defaultPathDefTuple.minWidth / 2),
        )
        self.draftLine = None
        self.m45Rotate = QTransform().rotate(-45)
        self._newPin = None
        self.newPinTuple = None
        self.newLabelTuple = None
        self.newLabel = None
        self.newRect = None
        self.newPolygon = None
        self.arrayViaTuple = None
        self._singleVia = None
        self.arrayVia = None
        self.polygonGuideLine = None
        self._newRuler = None
        self._newCutLine = None
        self._selectionRectItem = None
        self.rulersSet = set()
        self.rulerFont = self.setRulerFont(12 * fabproc.dbu)
        self.rulerFont.setKerning(False)
        self.rulerTickGap = fabproc.dbu
        self.rulerTickLength = 10
        self.rulerWidth = 2
        self.setItemIndexMethod(QGraphicsScene.ItemIndexMethod.BspTreeIndex)
        self._draftPen = QPen(Qt.PenStyle.DashLine)
        self._draftPen.setColor(QColor(0, 150, 0))
        self._draftPen.setWidth(2)
        self._draftPen.setCosmetic(True)
        self.setMinimumRenderSize(0.2)
        self.newAlignLine = None
        self.alignLineFinished.connect(alignToLine)

    @property
    def drawMode(self):
        return any(
            (
                self.editModes.drawPath,
                self.editModes.drawPin,
                self.editModes.drawArc,
                self.editModes.drawPolygon,
                self.editModes.drawRect,
                self.editModes.drawCircle,
                self.editModes.drawRuler,
            )
        )

    # Order of drawing
    # 1. Rect
    # 2. Path
    # 3. Pin
    # 4. Label
    # 5. Via/Contact
    # 6. Polygon
    # 7. Add instance
    # 8. select item/s
    # 9. rotate item/s

    @staticmethod
    def toLayoutCoord(point: QPoint) -> QPointF:
        """
        Converts a point in scene coordinates to layout coordinates by dividing it to
        fabproc.dbu.
        """
        scale = fabproc.dbu if fabproc else 1000
        if isinstance(point, QPointF):
            returnPoint = point / scale
        elif isinstance(point, QPoint):
            returnPoint = point.toPointF() / scale
        else:
            returnPoint = QPointF(0.0, 0.0)
        return returnPoint

    @staticmethod
    def toSceneCoord(point: Union[QPoint | QPointF]) -> QPoint:
        """
        Converts a point in layout coordinates to scene coordinates by multiplying it with
        fabproc.dbu.
        """
        point *= fabproc.dbu
        return point.toPoint() if isinstance(point, QPointF) else point

    def mousePressEvent(self, event: QGraphicsSceneMouseEvent) -> None:
                        
        self.mousePressLoc = event.scenePos().toPoint()
        if self.editModes.cutShape:
            self.startCutLine()
        super().mousePressEvent(event)

    def mouseMoveEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        """
        Handle the mouse move event.

        Args:
            mouse_event (QGraphicsSceneMouseEvent): The mouse event object.

        Returns:
            None
        """
        # Get the current mouse position
        self.mouseMoveLoc = event.scenePos().toPoint()
        # Call the parentW class's mouseMoveEvent method
        super().mouseMoveEvent(event)

        # Handle drawing path mode
        if self.editModes.drawPath and self.newPath is not None:
            self.newPath.draftLine = QLineF(
                self.newPath.draftLine.p1(), self.mouseMoveLoc
            )
        elif (self.editModes.drawRect or self.editModes.cutShape) and self.newRect:
            self.newRect.end = self.mouseMoveLoc
        # Handle drawing pin mode with no new pin
        elif self.editModes.drawPin:
            if self._newPin is not None:
                self._newPin.end = self.mouseMoveLoc
        elif self.editModes.addLabel and self.newLabel is not None:
            self.newLabel.start = self.mouseMoveLoc
        elif self.editModes.addInstance and self.newInstance is not None:
            self.newInstance.setPos(self.mouseMoveLoc - self.newInstance.start)
        # Handle drawing polygon mode
        elif self.editModes.drawPolygon and self.newPolygon is not None:
            self.polygonGuideLine.setLine(
                QLineF(self.newPolygon.points[-1], self.mouseMoveLoc)
            )
        elif self.editModes.drawRuler and self._newRuler is not None:
            self._newRuler.draftLine = QLineF(
                self._newRuler.draftLine.p1(), self.mouseMoveLoc
            )
        # Handle adding via mode with array via tuple
        elif self.editModes.addVia and self.arrayVia is not None:
            self.arrayVia.setPos(self.mouseMoveLoc - self.arrayVia.start)

        elif self.editModes.stretchItem and self.stretchPathItem is not None:
            self.stretchPathItem.draftLine = QLineF(
                self.stretchPathItem.draftLine.p1(), self.mouseMoveLoc
            )
        elif self.editModes.cutShape and self._newCutLine is not None:
            self._newCutLine.draftLine = QLineF(
                self._newCutLine.draftLine.p1(), self.mouseMoveLoc
            )
        elif self.newAlignLine and self.editModes.alignItems:
            self.newAlignLine.draftLine = QLineF(
                self.newAlignLine.draftLine.p1(), self.mouseMoveLoc
            )

        # Calculate the cursor position in layout units
        cursorPosition = self.toLayoutCoord(self.mouseMoveLoc - self.origin)

        # Show the cursor position in the status line
        self.statusLine.showMessage(
            f"Cursor Position: ({cursorPosition.x():.3f}u, {cursorPosition.y():.3f}u)"
        )
        self.messageLine.setText(self.messages[self.editModes.mode()])

    # def mouseReleaseEvent(self, mouse_event: QGraphicsSceneMouseEvent) -> None:
    #     super().mouseReleaseEvent(mouse_event)
    #     if mouse_event.button() == Qt.LeftButton:
    #         self.mouseReleaseLoc = mouse_event.scenePos().toPoint()
    #         self._handleMouseRelease(self.mouseReleaseLoc, mouse_event.button())

    def _handleMouseRelease(self, mousePos: QPoint, button: Qt.MouseButton) -> None:
        try:
            if self.editModes.drawPath:
                self.drawLayoutPath()
            elif self.editModes.drawRect:
                self.drawLayoutRect()
            elif self.editModes.drawPin:
                self.drawLayoutPin()
            elif self.editModes.addLabel:
                self.addLayoutLabel()
            elif self.editModes.addInstance:
                self.addLayoutInstance()
            elif self.editModes.drawPolygon:
                self.drawLayoutPolygon()
            elif self.editModes.drawRuler:
                self.drawLayoutRuler()
            elif self.editModes.cutShape:
                self.finishCutLine()
            elif self.editModes.addVia:
                self.addLayoutViaArray()
            elif self.editModes.changeOrigin:
                self.origin: QPoint = mousePos
            elif self.editModes.alignItems:
                self._handleAlignItemLine(mousePos)
            elif self.editModes.rotateItem:
                self.editorWindow.messageLine.setText("Rotate item")
                if self.selectedItems():
                    self.rotateSelectedItems(mousePos)
        except Exception as e:
            self.logger.error(f"mouse release error: {e}")

    def addLayoutViaArray(self):
        if self.arrayVia is not None:
            self.arrayViaTuple = None
            self.arrayVia = None

        singleVia = lshp.layoutVia(
            QPoint(0, 0),
            *self.arrayViaTuple.singleViaTuple,
        )
        self.arrayVia = lshp.layoutViaArray(
            self.mouseReleaseLoc,
            singleVia,
            self.arrayViaTuple.xs,
            self.arrayViaTuple.ys,
            self.arrayViaTuple.xnum,
            self.arrayViaTuple.ynum,
        )
        self.addUndoStack(self.arrayVia)

    def drawLayoutRuler(self):
        if self._newRuler:
            if self._newRuler.draftLine.isNull():
                self.undoStack.removeLastCommand()
            self._newRuler = None
        self._newRuler = lshp.layoutRuler(
            QLineF(self.mouseReleaseLoc, self.mouseReleaseLoc),
            width=self.rulerWidth,
            tickGap=self.rulerTickGap,
            tickLength=self.rulerTickLength,
            tickFont=self.rulerFont,
        )
        self.addUndoStack(self._newRuler)

    def finishCutLine(self):
        if self._newCutLine is None:
            return
        if self._newCutLine.draftLine.isNull():
            self.undoStack.removeLastCommand()
        else:
            if self.selectedItemsSet:
                line = QLineF(*self._newCutLine.sceneEndPoints)
                for item in self.selectedItemsSet:
                    if isinstance(item, lshp.layoutRect):
                        self._splitRect(item, line)
                    elif isinstance(item, lshp.layoutPath):
                        self._splitPath(item, line)
                    elif isinstance(item, lshp.layoutPolygon):
                        self._splitPolygon(item, line)
                self.removeItem(self._newCutLine)
        self._newCutLine = None

    def _splitPolygonalShape(self, item, line: QLineF):
        """Split a polygonal shape (rectangle or polygon) at intersections with the cut line."""
        # Get points based on item type
        if hasattr(item, 'rect'):  # Rectangle
            rect = item.rect
            points = [
                rect.topLeft(),
                rect.topRight(),
                rect.bottomRight(),
                rect.bottomLeft(),
            ]
            shape_name = "Rectangle"
        else:  # Polygon
            points = item.points
            shape_name = "Polygon"
        
        intersections = []

        # Find all intersection points with shape edges
        for i in range(len(points)):
            edge = QLineF(points[i], points[(i + 1) % len(points)])
            result = line.intersects(edge)
            if result[0] == QLineF.IntersectionType.BoundedIntersection:
                intersections.append((i, result[1]))

        # Split shape if we have exactly 2 intersections
        if len(intersections) == 2:
            idx1, p1 = intersections[0]
            idx2, p2 = intersections[1]

            # Create two new polygons
            poly1_points = [p1]
            for i in range(idx1 + 1, idx2 + 1):
                poly1_points.append(points[i])
            poly1_points.append(p2)

            poly2_points = [p2]
            for i in range(idx2 + 1, len(points)):
                poly2_points.append(points[i])
            for i in range(0, idx1 + 1):
                poly2_points.append(points[i])
            poly2_points.append(p1)

            poly1 = lshp.layoutPolygon(poly1_points, item.layer)
            poly2 = lshp.layoutPolygon(poly2_points, item.layer)

            self.undoStack.beginMacro(f"Split {shape_name}")
            self.deleteUndoStack(item)
            self.addUndoStack(poly1)
            self.addUndoStack(poly2)
            self.undoStack.endMacro()

    def _splitRect(self, item, line: QLineF):
        """Split a rectangle - delegates to unified polygonal shape splitter."""
        self._splitPolygonalShape(item, line)

    def _splitPolygon(self, item, line):
        """Split a polygon - delegates to unified polygonal shape splitter."""
        self._splitPolygonalShape(item, line)

    def startCutLine(self):
        if self.selectedItemsSet:
            self._newCutLine = lshp.layoutLine(
                QLineF(self.mousePressLoc, self.mousePressLoc), self.snapGrid, 1
            )
            self.addUndoStack(self._newCutLine)
        else:
            self.editorWindow.messageLine.setText("No selected items")

    def _handleAlignItemLine(self, eventLoc: QPoint) -> None:
        if self.newAlignLine is None:
            from PySide6.QtWidgets import (
                QApplication,
            )

            alignDlg = [
                w
                for w in QApplication.topLevelWidgets()
                if isinstance(w, alignItemsDialogue)
                   and w.isVisible()
                   and w.scene == self
            ][0]
            if alignDlg.horizontalAlignButton.isChecked():
                self.newAlignLine = lshp.alignLine(
                    QLineF(eventLoc, eventLoc), 1, 0
                )  # horizontal
            else:
                self.newAlignLine = lshp.alignLine(QLineF(eventLoc, eventLoc), 1, 1)
            self.addUndoStack(self.newAlignLine)
        else:
            self.newAlignLine.draftLine = QLineF(
                self.newAlignLine.draftLine.p1(), eventLoc
            )
            self.alignLineFinished.emit(self.newAlignLine)
            # self.newAlignLine = None
            self.editModes.setMode("selectItem")

    def drawLayoutPolygon(self):
        if self.newPolygon is None:
            # Create a new polygon
            self.newPolygon = lshp.layoutPolygon(
                [self.mouseReleaseLoc, self.mouseReleaseLoc],
                self.selectEdLayer,
            )
            self.addUndoStack(self.newPolygon)
            # Create a guide line for the polygon
            self.polygonGuideLine = QGraphicsLineItem(
                QLineF(self.newPolygon.points[-2], self.newPolygon.points[-1])
            )
            self.polygonGuideLine.setPen(
                QPen(QColor(255, 255, 0), 0.1 * fabproc.dbu, Qt.PenStyle.DashLine)
            )
            self.polygonGuideLine.pen().setCosmetic(False)
            self.addUndoStack(self.polygonGuideLine)
        else:
            self.newPolygon.addPoint(self.mouseReleaseLoc)

    def addLayoutInstance(self):
        if self.newInstance is not None:
            self.newInstance = None
        if self.layoutInstanceTuple:
            self.newInstance = self.addNewInstance()
            self.addUndoStack(self.newInstance)
            self.newInstance.setPos(self.mouseReleaseLoc - self.newInstance.start)

    def addLayoutLabel(self):
        if self.newLabel is not None:
            self.newLabelTuple = None
            self.newLabel = None
        if self.newLabelTuple is not None:
            self.newLabel = lshp.layoutLabel(self.mouseReleaseLoc, *self.newLabelTuple)
            self.addUndoStack(self.newLabel)

    def drawLayoutPin(self):
        self.editorWindow.messageLine.setText("Pin mode.")
        # Create a new pin
        if self._newPin:
            if self._newPin.rect.isNull():
                self.undoStack.removeLastCommand()
            else:
                self.editModes.setMode("addLabel")
                self.newLabel = lshp.layoutLabel(
                    self.mouseReleaseLoc, *self.newLabelTuple
                )
                self._newPin.label = self.newLabel
                self.addUndoStack(self.newLabel)
            self._newPin = None
        self._newPin = lshp.layoutPin(
            self.mouseReleaseLoc, self.mouseReleaseLoc, *self.newPinTuple
        )
        self.addUndoStack(self._newPin)

    def drawLayoutRect(self):
        self.editorWindow.messageLine.setText("Rectangle mode.")
        # Create a new rectangle
        if self.newRect:
            if self.newRect.rect.isNull():
                self.removeItem(self.newRect)
                self.undoStack.removeLastCommand()
            self.newRect = None
        self.newRect = lshp.layoutRect(
            self.mouseReleaseLoc,
            self.mouseReleaseLoc,
            self.selectEdLayer,
        )
        self.addUndoStack(self.newRect)

    def drawLayoutPath(self):
        self.editorWindow.messageLine.setText("Path mode")
        if self.newPath:
            if self.newPath.draftLine.isNull():
                self.undoStack.removeLastCommand()
            self.newPath = None

            # Create a new path
        self.newPath = lshp.layoutPath(
            QLineF(self.mousePressLoc, self.mousePressLoc),
            self.newPathTuple.layer,
            self.newPathTuple.width,
            int(self.newPathTuple.startExtend),
            int(self.newPathTuple.endExtend),
            self.newPathTuple.pathMode,
        )
        self.newPath.name = self.newPathTuple.name
        self.addUndoStack(self.newPath)

    def addNewInstance(self) -> Union[lshp.layoutInstance, lshp.layoutPcell]:
        newInstance = self.instLayout(self.layoutInstanceTuple)
        if isinstance(newInstance, pcells.baseCell):
            dlg = ldlg.layoutInstanceDialogue(self.editorWindow)
            dlg.instanceLibName.setText(newInstance.libraryName)
            dlg.instanceCellName.setText(newInstance.cellName)
            dlg.instanceViewName.setText(newInstance.viewName)
            lineEditDict = self.extractPcellInstanceParameters(newInstance)
            if lineEditDict:
                dlg.pcellParamsGroup.show()
                for key, value in lineEditDict.items():
                    dlg.pcellParamsLayout.addRow(key, value)
            if dlg.exec() == QDialog.DialogCode.Accepted:
                instanceValuesDict = {}
                for key, value in lineEditDict.items():
                    instanceValuesDict[key] = value.text()
                if instanceValuesDict:
                    newInstance(*instanceValuesDict.values())
        self.layoutInstanceTuple = None
        return newInstance

    def instLayout(
            self, layoutInstanceTuple: ddef.viewItemTuple
    ) -> Union[lshp.layoutInstance, lshp.layoutPcell]:
        """Read a layout file and create layoutShape objects from it."""
        try:
            with layoutInstanceTuple.viewItem.viewPath.open("r") as temp:
                decodedData = json.load(temp)

            # Common instance setup
            def setup_instance(instance):
                instance.libraryName = layoutInstanceTuple.libraryItem.libraryName
                instance.cellName = layoutInstanceTuple.cellItem.cellName
                instance.viewName = layoutInstanceTuple.viewItem.viewName
                self.itemCounter += 1
                instance.counter = self.itemCounter
                instance.instanceName = f"I{self.itemCounter}"
                return instance

            viewType = layoutInstanceTuple.viewItem.viewType

            if (
                    viewType == "layout"
                    and decodedData
                    and decodedData[0].get("viewType") == "layout"
            ):
                factory_create = lj.layoutItems(self).create
                # valid_items = filter(
                #     lambda item: isinstance(item, dict) and item.get(
                #         "type") in self.LAYOUT_SHAPES, decodedData[2:])
                instanceShapes = [factory_create(item) for item in decodedData[2:]]
                return setup_instance(lshp.layoutInstance(instanceShapes))

            elif (
                    viewType == "pcell"
                    and len(decodedData) > 1
                    and decodedData[0].get("cellView") == "pcell"
            ):
                pcellInstance = eval(f"pcells.{decodedData[1]['reference']}()")
                return setup_instance(pcellInstance)

        except Exception:
            pass

    def findScenelayoutCellSet(self) -> set[lshp.layoutInstance]:
        """
        Find all the symbols on the scene as a set.
        """
        return {item for item in self.items() if isinstance(item, lshp.layoutInstance)}

    def saveLayoutCell(self, filePathObj: pathlib.Path) -> None:
        """Save the layout cell to a JSON file.

        Args:
            filePathObj (pathlib.Path): Path object for the output file

        Raises:
            IOError: If there are issues writing to the file
            ValueError: If the layout data is invalid
        """

        def safeJsonWrite(file_obj, data: list) -> None:
            """Write JSON data with optimized settings.

            Args:
                file_obj: File object to write to
                data: Data to be written

            Raises:
                JSONEncodeError: If JSON encoding fails
            """
            json.dump(
                data,
                file_obj,
                cls=layenc.layoutEncoder,
                separators=(",", ":"),  # Minimize JSON size
                check_circular=False,  # Optimize for non-circular references
            )

        try:
            # Create parentW directory if it doesn't exist
            filePathObj.parent.mkdir(parents=True, exist_ok=True)

            # Prepare data before file operation
            self.itemsRefSet = set(self.items())
            topLevelItems = [
                item
                for item in self.itemsRefSet
                if item.parentItem() is None
                   and isinstance(item, tuple(self.LAYOUT_SHAPES))
            ]
            layoutData = [
                {"viewType": "layout"},
                {"snapGrid": (self.majorGrid, self.snapGrid)},
                *topLevelItems,
            ]

            # Use temporary file for atomic write
            temp_path = filePathObj.with_suffix(".tmp")
            try:
                with temp_path.open(mode="w", buffering=65536) as f:  # 64KB buffer
                    safeJsonWrite(f, layoutData)

                # Atomic rename for safer file writing
                temp_path.replace(filePathObj)

            finally:
                # Clean up temp file if it still exists
                if temp_path.exists():
                    temp_path.unlink()

            self.logger.info(
                f"Saved layout to {self.editorWindow.cellName}:{self.editorWindow.viewName}"
            )

        except ValueError as e:
            self.logger.error(f"Invalid layout data: {str(e)}")
            raise

        except (IOError, TypeError) as e:
            self.logger.error(f"Failed to save layout to {filePathObj}: {str(e)}")
            raise

        except Exception as e:
            self.logger.error(f"Unexpected error while saving layout: {str(e)}")
            raise

    def _exportCell(
            self, export_dir: pathlib.Path, unit: float, precision: float, dbu: int, format_type: str = "GDS"
    ):
        """Export cell in the specified format (GDS or OAS)."""
        format_upper = format_type.upper()
        file_extension = ".gds" if format_upper == "GDS" else ".oas"
        export_path = export_dir / f"{self.cellName}{file_extension}"
        
        try:
            # reprocess the layout to get the layout positions right.
            topLevelItems = [
                item
                for item in self.itemsRefSet
                if item.parentItem() is None
                   and isinstance(item, tuple(self.LAYOUT_SHAPES))
            ]

            decodedData = json.loads(
                json.dumps(topLevelItems, cls=layenc.layoutEncoder)
            )

            layoutItems = [lj.layoutItems(self).create(item) for item in decodedData]

            exportObj = gdse.gdsExporter(self.cellName, layoutItems, export_path)
            exportObj.unit = unit
            exportObj.precision = precision
            exportObj.dbu = dbu

            self.logger.info(f"{format_upper} Export started.")
            with self.measureDuration():
                if format_upper == "OAS":
                    exportObj.oasExportThreaded(self.appMainW.threadPool)
                else:
                    exportObj.gdsExportThreaded(self.appMainW.threadPool)
            self.logger.info(f"{format_upper} Export finished.")

        except ValueError as e:
            self.logger.error(f"Invalid layout data: {str(e)}")
            raise

        except (IOError, TypeError) as e:
            self.logger.error(f"Failed to export layout to {export_path}: {str(e)}")
            raise

        except Exception as e:
            self.logger.error(f"Unexpected error while exporting layout: {str(e)}")
            raise

    def exportCellGDS(
            self, gdsExportDir: pathlib.Path, gdsUnit: float, gdsPrecision: float, dbu: int
    ):
        """Export cell as GDS format."""
        self._exportCell(gdsExportDir, gdsUnit, gdsPrecision, dbu, "GDS")

    def exportCellOAS(
            self, oasExportDir: pathlib.Path, gdsUnit: float, gdsPrecision: float, dbu: int
    ):
        """Export cell as OAS format."""
        self._exportCell(oasExportDir, gdsUnit, gdsPrecision, dbu, "OAS")

    def loadDesign(self, filePathObj: pathlib.Path) -> bool:
        """Load the layout cell from the given JSON file."""
        try:
            with filePathObj.open("rb") as file:
                decodedData = orjson.loads(file.read())
            if len(decodedData) < 2 or decodedData[0].get("viewType") != "layout":
                self.logger.error("Invalid file type.")
                return False
            with self.measureDuration():
                self.editorWindow.configureGridSettings(
                    decodedData[1].get("snapGrid", (self.majorGrid, self.snapGrid))
                )
                if len(decodedData) > 2:
                    self.createLayoutItems(decodedData[2:])
            self.itemsRefSet = set(self.items())
            return True
        except orjson.JSONDecodeError:
            self.logger.error("Invalid file format.")
            return False
        except Exception as e:
            self.logger.error(f"Unexpected error loading layout: {e}")
            return False

    def createLayoutItems(self, decoded_data: List[Dict[str, Any]]) -> None:
        if not decoded_data:
            return

        factory_create = lj.layoutItems(self).create
        for item in decoded_data:
            if isinstance(item, dict):
                try:
                    self.addItem(factory_create(item))
                except Exception:
                    pass

    def deleteSelectedItems(self):
        for item in self.selectedItems():
            # if pin is to be deleted, the associated label should be also deleted.
            if isinstance(item, lshp.layoutPin) and item.label is not None:
                undoCommand = us.deleteShapeUndo(self, item.label)
                self.undoStack.push(undoCommand)
        super().deleteSelectedItems()

    def viewObjProperties(self):
        """
        Display the properties of the selected object.
        """
        try:
            selectedItems = [
                item for item in self.selectedItems() if item.parentItem() is None
            ]
            if selectedItems is not None:
                for item in selectedItems:
                    match type(item):
                        case lshp.layoutRect:
                            self.layoutRectProperties(item)
                        case lshp.layoutPin:
                            self.layoutPinProperties(item)
                        case lshp.layoutLabel:
                            self.layoutLabelProperties(item)
                        case lshp.layoutPath:
                            self.layoutPathProperties(item)
                        case lshp.layoutViaArray:
                            self.layoutViaProperties(item)
                        case lshp.layoutPolygon:
                            self.layoutPolygonProperties(item)
                        case lshp.layoutInstance:
                            self.layoutInstanceProperties(item, False)
                        # case _:
                        # if item.__class__.__bases__[0] == pcells.baseCell:
                        #     self.layoutInstanceProperties(item, True)
                        case _:
                            if isinstance(item, pcells.baseCell):
                                self.layoutInstanceProperties(item, True)

        except Exception as e:
            self.logger.error(f"{type(item)} property editor error: {e}")

    def layoutPolygonProperties(self, item):
        pointsTupleList = [self.toLayoutCoord(point).toTuple() for point in item.points]

        dlg = ldlg.layoutPolygonProperties(self.editorWindow, pointsTupleList)
        dlg.polygonLayerCB.addItems(
            [f"{item.name} [{item.purpose}]" for item in laylyr.pdkAllLayers]
        )
        dlg.polygonLayerCB.setCurrentText(f"{item.layer.name} [{item.layer.purpose}]")

        if dlg.exec() == QDialog.DialogCode.Accepted:
            newLayer = laylyr.pdkAllLayers[dlg.polygonLayerCB.currentIndex()]
            tempPoints = []
            for i in range(dlg.tableWidget.rowCount()):
                xcoor = dlg.tableWidget.item(i, 1).text()
                ycoor = dlg.tableWidget.item(i, 2).text()
                if xcoor != "" and ycoor != "":
                    tempPoints.append(
                        self.toSceneCoord(QPointF(float(xcoor), float(ycoor)))
                    )
            newPoints = tempPoints
            newPolygon = lshp.layoutPolygon(newPoints, newLayer)
            self.undoStack.push(us.addDeleteShapeUndo(self, newPolygon, item))

    def layoutRectProperties(self, item):
        dlg = ldlg.layoutRectProperties(self.editorWindow)
        dlg.rectLayerCB.addItems(
            [f"{item.name} [{item.purpose}]" for item in laylyr.pdkAllLayers]
        )
        dlg.rectLayerCB.setCurrentText(f"{item.layer.name} [{item.layer.purpose}]")
        dlg.rectWidthEdit.setText(str(item.width / fabproc.dbu))
        dlg.rectHeightEdit.setText(str(item.height / fabproc.dbu))
        dlg.topLeftEditX.setText(str(item.rect.topLeft().x() / fabproc.dbu))
        dlg.topLeftEditY.setText(str(item.rect.topLeft().y() / fabproc.dbu))
        if dlg.exec() == QDialog.DialogCode.Accepted:
            layer = laylyr.pdkAllLayers[dlg.rectLayerCB.currentIndex()]
            newRect = lshp.layoutRect(QPoint(0, 0), QPoint(0, 0), layer)
            newRect.rect = QRectF(
                float(dlg.topLeftEditX.text()) * fabproc.dbu,
                float(dlg.topLeftEditY.text()) * fabproc.dbu,
                float(dlg.rectWidthEdit.text()) * fabproc.dbu,
                float(dlg.rectHeightEdit.text()) * fabproc.dbu,
            ).toRect()
            self.undoStack.push(us.addDeleteShapeUndo(self, newRect, item))

    def layoutViaProperties(self, item):
        dlg = ldlg.layoutViaProperties(self.editorWindow)
        if item.xnum == 1 and item.ynum == 1:
            dlg.singleViaRB.setChecked(True)
            dlg.singleViaClicked()
            dlg.singleViaNamesCB.addItems(fabproc.processViaNames)
            dlg.singleViaNamesCB.setCurrentText(item.via.viaDefTuple.name)
            dlg.singleViaWidthEdit.setText(str(item.width / fabproc.dbu))
            dlg.singleViaHeightEdit.setText(str(item.via.height / fabproc.dbu))
        else:
            dlg.arrayViaRB.setChecked(True)
            dlg.arrayViaClicked()
            dlg.arrayViaNamesCB.addItems(fabproc.processViaNames)
            dlg.arrayViaNamesCB.setCurrentText(item.via.viaDefTuple.name)
            dlg.arrayViaWidthEdit.setText(str(item.via.width / fabproc.dbu))
            dlg.arrayViaHeightEdit.setText(str(item.via.height / fabproc.dbu))
            dlg.arrayXspacingEdit.setText(str(item.xs / fabproc.dbu))
            dlg.arrayYspacingEdit.setText(str(item.xs / fabproc.dbu))
            dlg.arrayXNumEdit.setText(str(item.xnum))
            dlg.arrayYNumEdit.setText(str(item.ynum))
        dlg.startXEdit.setText(str(self.toLayoutCoord(item.mapToScene(item.start)).x()))
        dlg.startYEdit.setText(str(self.toLayoutCoord(item.mapToScene(item.start)).y()))
        if dlg.exec() == QDialog.DialogCode.Accepted:
            startX = int(float(dlg.startXEdit.text()))
            startY = int(float(dlg.startYEdit.text()))
            start = self.toSceneCoord(QPoint(startX, startY))
            if dlg.singleViaRB.isChecked():
                selViaDefTuple = fabproc.processVias[
                    fabproc.processViaNames.index(dlg.singleViaNamesCB.currentText())
                ]
                singleViaTuple = ddef.singleViaTuple(
                    selViaDefTuple,
                    float(dlg.singleViaWidthEdit.text().strip()) * fabproc.dbu,
                    float(dlg.singleViaHeightEdit.text().strip()) * fabproc.dbu,
                )
                arrayViaTuple = ddef.arrayViaTuple(
                    singleViaTuple,
                    float(selViaDefTuple.minSpacing) * fabproc.dbu,
                    float(selViaDefTuple.minSpacing) * fabproc.dbu,
                    1,
                    1,
                )
            else:
                selViaDefTuple = [
                    viaDefTuple
                    for viaDefTuple in fabproc.processVias
                    if viaDefTuple.name == dlg.arrayViaNamesCB.currentText()
                ][0]

                singleViaTuple = ddef.singleViaTuple(
                    selViaDefTuple,
                    float(dlg.arrayViaWidthEdit.text().strip()) * fabproc.dbu,
                    float(dlg.arrayViaHeightEdit.text().strip()) * fabproc.dbu,
                )
                arrayViaTuple = ddef.arrayViaTuple(
                    singleViaTuple,
                    float(dlg.arrayXspacingEdit.text().strip()) * fabproc.dbu,
                    float(dlg.arrayYspacingEdit.text().strip()) * fabproc.dbu,
                    int(float(dlg.arrayXNumEdit.text().strip())),
                    int(float(dlg.arrayYNumEdit.text().strip())),
                )
            singleVia = lshp.layoutVia(
                QPoint(0, 0),
                *arrayViaTuple.singleViaTuple,
            )
            arrayVia = lshp.layoutViaArray(
                start,
                singleVia,
                arrayViaTuple.xs,
                arrayViaTuple.ys,
                arrayViaTuple.xnum,
                arrayViaTuple.ynum,
            )
            self.undoStack.push(us.addDeleteShapeUndo(self, arrayVia, item))

    def layoutPathProperties(self, item):
        dlg = ldlg.layoutPathPropertiesDialog(self.editorWindow)
        match item.mode:
            case 0:
                dlg.manhattanButton.setChecked(True)
            case 1:
                dlg.diagonalButton.setChecked(True)
            case 2:
                dlg.anyButton.setChecked(True)
            case 3:
                dlg.horizontalButton.setChecked(True)
            case 4:
                dlg.verticalButton.setChecked(True)

        dlg.pathLayerCB.addItems(fabproc.processPathNames)

        currentIndex = next(
            (
                i
                for i, pathDefTuple in enumerate(fabproc.processPaths)
                if pathDefTuple.layer.name == item.layer.name
            ),
            0,
        )
        dlg.pathLayerCB.setCurrentIndex(currentIndex)

        dlg.pathWidth.setText(str(item.width / fabproc.dbu))
        dlg.pathNameEdit.setText(item.name)
        roundingFactor = len(str(fabproc.dbu)) - 1
        dlg.startExtendEdit.setText(
            str(round(item.startExtend / fabproc.dbu, roundingFactor))
        )
        dlg.endExtendEdit.setText(
            str(round(item.endExtend / fabproc.dbu, roundingFactor))
        )
        dlg.p1PointEditX.setText(
            str(round(item.draftLine.p1().x() / fabproc.dbu, roundingFactor))
        )
        dlg.p1PointEditY.setText(
            str(round(item.draftLine.p1().y() / fabproc.dbu, roundingFactor))
        )
        dlg.p2PointEditX.setText(
            str(round(item.draftLine.p2().x() / fabproc.dbu, roundingFactor))
        )
        dlg.p2PointEditY.setText(
            str(round(item.draftLine.p2().y() / fabproc.dbu, roundingFactor))
        )
        dlg.angleEdit.setText(str(item.angle))
        if dlg.exec() == QDialog.DialogCode.Accepted:
            item.name = dlg.pathNameEdit.text()
            pathDefTuple = fabproc.processPaths[dlg.pathLayerCB.currentIndex()]
            width = fabproc.dbu * float(dlg.pathWidth.text())
            startExtend = fabproc.dbu * float(dlg.startExtendEdit.text())
            endExtend = fabproc.dbu * float(dlg.endExtendEdit.text())
            p1 = self.toSceneCoord(
                QPointF(
                    float(dlg.p1PointEditX.text()),
                    float(dlg.p1PointEditY.text()),
                )
            )
            p2 = self.toSceneCoord(
                QPointF(
                    float(dlg.p2PointEditX.text()),
                    float(dlg.p2PointEditY.text()),
                )
            )
            draftLine = QLineF(p1, p2)
            newPath = lshp.layoutPath(
                draftLine, pathDefTuple.layer, width, startExtend, endExtend
            )
            newPath.angle = int(float(dlg.angleEdit.text()))
            self.undoStack.push(us.addDeleteShapeUndo(self, newPath, item))

    def layoutLabelProperties(self, item):
        dlg = ldlg.layoutLabelProperties(self.editorWindow)
        dlg.labelName.setText(item.labelText)
        dlg.labelLayerCB.addItems(
            [f"{layer.name} [{layer.purpose}]" for layer in laylyr.pdkTextLayers]
        )
        dlg.labelLayerCB.setCurrentText(f"{item.layer.name} [{item.layer.purpose}]")
        dlg.familyCB.setCurrentText(item.fontFamily)
        dlg.fontStyleCB.setCurrentText(item.fontStyle)
        dlg.labelHeightCB.setCurrentText(str(int(float(item.fontHeight))))
        dlg.labelAlignCB.setCurrentText(item.labelAlign)
        dlg.labelOrientCB.setCurrentText(item.labelOrient)
        start = item.mapToScene(item.start)
        dlg.labelTopLeftX.setText(str(self.toLayoutCoord(start).x()))
        dlg.labelTopLeftY.setText(str(self.toLayoutCoord(start).y()))

        if dlg.exec() == QDialog.DialogCode.Accepted:
            labelName = dlg.labelName.text()
            labelStartX = float(dlg.labelTopLeftX.text())
            labelStartY = float(dlg.labelTopLeftY.text())

            labelStart = self.toSceneCoord(QPointF(labelStartX, labelStartY))
            labelLayerName = dlg.labelLayerCB.currentText().split()[0]
            labelLayer = [
                item for item in laylyr.pdkTextLayers if item.name == labelLayerName
            ][0]
            fontFamily = dlg.familyCB.currentText()
            fontStyle = dlg.fontStyleCB.currentText()
            fontHeight = dlg.labelHeightCB.currentText()
            labelAlign = dlg.labelAlignCB.currentText()
            labelOrient = dlg.labelOrientCB.currentText()
            newLabelTuple = ddef.layoutLabelTuple(
                labelName,
                fontFamily,
                fontStyle,
                fontHeight,
                labelAlign,
                labelOrient,
                labelLayer,
            )
            newLabel = lshp.layoutLabel(labelStart, *newLabelTuple)
            self.undoStack.push(us.addDeleteShapeUndo(self, newLabel, item))

    def layoutPinProperties(self, item):
        dlg = ldlg.layoutPinProperties(self.editorWindow)
        dlg.pinName.setText(item.pinName)
        dlg.pinDir.setCurrentText(item.pinDir)
        dlg.pinType.setCurrentText(item.pinType)

        dlg.pinLayerCB.addItems(
            [
                f"{pinLayer.name} [{pinLayer.purpose}]"
                for pinLayer in laylyr.pdkPinLayers
            ]
        )
        dlg.pinLayerCB.setCurrentText(f"{item.layer.name} [{item.layer.purpose}]")
        dlg.pinBottomLeftX.setText(str(item.mapToScene(item.start).x() / fabproc.dbu))
        dlg.pinBottomLeftY.setText(str(item.mapToScene(item.start).y() / fabproc.dbu))
        dlg.pinTopRightX.setText(str(item.mapToScene(item.end).x() / fabproc.dbu))
        dlg.pinTopRightY.setText(str(item.mapToScene(item.end).y() / fabproc.dbu))
        if dlg.exec() == QDialog.DialogCode.Accepted:
            pinName = dlg.pinName.text()
            pinDir = dlg.pinDir.currentText()
            pinType = dlg.pinType.currentText()
            labelText = dlg.pinName.text()
            start = self.snapToGrid(
                self.toSceneCoord(
                    QPointF(
                        float(dlg.pinBottomLeftX.text()),
                        float(dlg.pinBottomLeftY.text()),
                    )
                )
            )
            end = self.snapToGrid(
                self.toSceneCoord(
                    QPointF(
                        float(dlg.pinTopRightX.text()),
                        float(dlg.pinTopRightY.text()),
                    )
                )
            )
            pinLayer = laylyr.pdkPinLayers[dlg.pinLayerCB.currentIndex()]
            newPin = lshp.layoutPin(start, end, pinName, pinDir, pinType, pinLayer)
            newLabel = lshp.layoutLabel(
                item.label.start,
                labelText,
                item.label.fontFamily,
                item.label.fontStyle,
                item.label.fontHeight,
                item.label.labelAlign,
                item.label.labelOrient,
                item.label.layer,
            )
            newPin.label = newLabel

            undoCommandsList = [
                us.addDeleteShapeUndo(self, newPin, item),
                us.addDeleteShapeUndo(self, newLabel, item.label),
            ]
            self.addUndoMacroStack(undoCommandsList, "pin/label edit")
            # self.undoStack.beginMacro('pin/label edit')
            # self.undoStack.push(us.addDeleteShapeUndo(self, newPin, item))
            # self.undoStack.push(us.addDeleteShapeUndo(self, newLabel, item.label))
            # self.undoStack.endMacro()

    def layoutInstanceProperties(
            self, item: Union[lshp.layoutInstance, lshp.layoutPcell], pcell: bool = False
    ):
        libraryModel = lmview.layoutViewsModel(
            self.editorWindow.libraryDict, self.editorWindow.layoutViews
        )
        dlg = ldlg.layoutInstancePropertiesDialogue(self.editorWindow)
        if pcell:
            dlg.pcellParamsGroup.show()
            lineEditDict = self.extractPcellInstanceParameters(item)
            if lineEditDict:
                dlg.pcellParamsGroup.show()
                for key, value in lineEditDict.items():
                    dlg.pcellParamsLayout.addRow(key, value)
        libItem = libm.getLibItem(libraryModel, item.libraryName)
        cellItem = libm.getCellItem(libItem, item.cellName)
        viewItem = libm.getViewItem(cellItem, item.viewName)
        instanceTuple = ddef.viewItemTuple(libItem, cellItem, viewItem)
        dlg.instanceLibName.setText(item.libraryName)
        libNameCompleter = QCompleter(libraryModel.listLibraries())
        libNameCompleter.setCaseSensitivity(Qt.CaseInsensitive)
        dlg.instanceLibName.setCompleter(libNameCompleter)
        # create cell name completer list.
        dlg.instanceLibName.editingFinished.connect(
            lambda: self.cellNameComplete(
                dlg, libraryModel.listLibraryCells(dlg.instanceLibName.text())
            )
        )
        dlg.instanceCellName.setText(item.cellName)
        # create view name completer list
        dlg.instanceCellName.editingFinished.connect(
            lambda: self.viewNameComplete(
                dlg,
                libraryModel.listCellViews(
                    dlg.instanceLibName.text(),
                    dlg.instanceCellName.text(),
                    ["layout", "pcell"],
                ),
            )
        )
        dlg.instanceViewName.setText(item.viewName)
        dlg.instanceViewName.editingFinished.connect(
            lambda: self.changePcellParameterFields(dlg, libraryModel, instanceTuple)
        )
        dlg.instanceNameEdit.setText(item.instanceName)

        dlg.xEdit.setText(str(item.scenePos().x() / fabproc.dbu))
        dlg.yEdit.setText(str(item.scenePos().y() / fabproc.dbu))

        if dlg.exec() == QDialog.DialogCode.Accepted:
            libraryName = dlg.instanceLibName.text().strip()
            cellName = dlg.instanceCellName.text().strip()
            viewName = dlg.instanceViewName.text().strip()
            instanceName = dlg.instanceNameEdit.text().strip()
            libItem = libm.getLibItem(libraryModel, libraryName)
            cellItem = libm.getCellItem(libItem, cellName)
            viewItem = libm.getViewItem(cellItem, viewName)

            layoutInstanceTuple = ddef.viewItemTuple(libItem, cellItem, viewItem)
            newLayoutInstance = self.instLayout(layoutInstanceTuple)
            newLayoutInstance.instanceName = instanceName
            lineEditDict = ldlg.formDictionary(
                dlg.pcellParamsLayout
            ).extractDictFormLayout()
            instanceValuesDict = {}
            if lineEditDict:
                for key, value in lineEditDict.items():
                    instanceValuesDict[key] = value.text()
            if instanceValuesDict:
                newLayoutInstance(**instanceValuesDict)
            newLayoutInstance.setPos(
                QPoint(
                    self.snapToBase(
                        float(dlg.xEdit.text()) * fabproc.dbu, self.snapTuple[0]
                    ),
                    self.snapToBase(
                        float(dlg.yEdit.text()) * fabproc.dbu, self.snapTuple[1]
                    ),
                )
            )
            self.undoStack.push(us.addDeleteShapeUndo(self, newLayoutInstance, item))

    def changePcellParameterFields(
            self,
            dlg: ldlg.layoutInstancePropertiesDialogue,
            libraryModel: lmview.layoutViewsModel,
            instanceTuple: ddef.viewItemTuple,
    ):
        """Update the PCell parameter fields based on the selected instance tuple."""
        # Get the new item tuple
        lib_item = libm.getLibItem(libraryModel, dlg.instanceLibName.text().strip())
        cell_item = libm.getCellItem(lib_item, dlg.instanceCellName.text().strip())
        view_item = libm.getViewItem(cell_item, dlg.instanceViewName.text().strip())
        new_item_tuple = ddef.viewItemTuple(lib_item, cell_item, view_item)

        # Check if the view item has a 'viewType' key
        if hasattr(new_item_tuple.viewItem, "viewType"):
            # Clear the pcell parameters layout
            self.clearLayout(dlg.pcellParamsLayout)

            newInstance = self.instLayout(new_item_tuple)

            # Check the view type
            if new_item_tuple.viewItem.viewType == "pcell":
                # Extract PCell instance parameters
                line_edit_dict = self.extractPcellInstanceParameters(newInstance)

                # Add PCell parameters to the layout
                if line_edit_dict:
                    dlg.pcellParamsGroup.show()
                    for key, value in line_edit_dict.items():
                        dlg.pcellParamsLayout.addRow(key, value)
            elif new_item_tuple.viewItem.viewType == "layout":
                # Hide the PCell parameters group
                dlg.pcellParamsGroup.hide()

    def extractPcellInstanceParameters(self, instance: lshp.layoutPcell) -> dict:
        initArgs = inspect.signature(instance.__class__.__init__).parameters
        argsUsed = [param for param in initArgs if (param != "self")]
        argDict = {arg: getattr(instance, arg) for arg in argsUsed}
        lineEditDict = {key: edf.shortLineEdit(value) for key, value in argDict.items()}
        return lineEditDict

    def clearLayout(self, layout):
        """Clear the layout by deleting all its children."""
        while layout.count():
            child = layout.takeAt(0)
            if child.widget():
                child.widget().deleteLater()
            elif child.layout():
                self.clearLayout(child.layout())

    def moveBySelectedItems(self):
        if self.selectedItems():
            dlg = pdlg.moveByDialogue(self.editorWindow)
            dlg.xEdit.setText("0.0")
            dlg.yEdit.setText("0.0")
            if dlg.exec() == QDialog.DialogCode.Accepted:
                for item in self.selectedItems():
                    item.moveBy(
                        self.snapToBase(
                            float(dlg.xEdit.text()) * fabproc.dbu, self.snapTuple[0]
                        ),
                        self.snapToBase(
                            float(dlg.yEdit.text()) * fabproc.dbu, self.snapTuple[1]
                        ),
                    )
                self.editorWindow.messageLine.setText(
                    f"Moved items by {dlg.xEdit.text()} and {dlg.yEdit.text()}"
                )
                self.editModes.setMode("selectItem")

    def copySelectedItems(self):
        selectedItems = [
            item for item in self.selectedItems() if item.parentItem() is None
        ]
        copyShapesList = []
        if selectedItems:
            for item in selectedItems:
                selectedItemJson = json.dumps(item, cls=layenc.layoutEncoder)
                itemCopyDict = json.loads(selectedItemJson)
                shape = lj.layoutItems(self).create(itemCopyDict)
                if shape is not None:
                    if isinstance(shape, (lshp.layoutInstance, lshp.layoutPcell)):
                        self.itemCounter += 1
                        shape.instanceName = f"I{self.itemCounter}"
                        shape.counter = int(self.itemCounter)
                    copyShapesList.append(shape)

            self.addListUndoStack(copyShapesList)
            self.selectedItemGroup = self.createItemGroup(copyShapesList)
            self.selectedItemGroup.setSelected(True)

    def deleteAllRulers(self):
        sceneRulerSet = {
            item for item in self.items() if isinstance(item, lshp.layoutRuler)
        }
        self.deleteListUndoStack(list(sceneRulerSet))
        # for ruler in self.rulersSet:
        #     undoCommand = us.deleteShapeUndo(self, ruler)
        #     self.undoStack.push(undoCommand)

    def goDownHier(self):
        if self.selectedItems():
            for item in self.selectedItems():
                if isinstance(item, lshp.layoutInstance):
                    dlg = fd.goDownHierDialogue(self.editorWindow)
                    libItem = libm.getLibItem(
                        self.editorWindow.libraryView.libraryModel, item.libraryName
                    )
                    cellItem = libm.getCellItem(libItem, item.cellName)
                    viewNames = [
                        cellItem.child(i).text()
                        for i in range(cellItem.rowCount())
                        # if cellItem.child(i).text() != item.viewName
                        if "layout" in cellItem.child(i).text()
                    ]
                    dlg.viewListCB.addItems(viewNames)
                    if dlg.exec() == QDialog.DialogCode.Accepted:
                        libItem = libm.getLibItem(
                            self.editorWindow.libraryView.libraryModel, item.libraryName
                        )
                        cellItem = libm.getCellItem(libItem, item.cellName)
                        viewItem = libm.getViewItem(
                            cellItem, dlg.viewListCB.currentText()
                        )
                        viewItemT = ddef.viewItemTuple(libItem, cellItem, viewItem)
                        openViewNameT = (
                            self.editorWindow.libraryView.openCellView(
                                viewItemT
                            )
                        )
                        if self.editorWindow.appMainW.openViews[openViewNameT]:
                            childWindow = self.editorWindow.appMainW.openViews[
                                openViewNameT
                            ]
                            childWindow.parentEditor = self.editorWindow
                            childWindow.parentObj = item
                            childWindow.layoutToolbar.addAction(childWindow.goUpAction)
                            if dlg.buttonId == 2:
                                childWindow.centralW.scene.readOnly = True

    def stretchPath(self, pathItem: lshp.layoutPath, stretchEnd: str):
        match stretchEnd:
            case "p2":
                self.stretchPathItem = lshp.layoutPath(
                    QLineF(pathItem.sceneEndPoints[0], pathItem.sceneEndPoints[1]),
                    pathItem.layer,
                    pathItem.width,
                    pathItem.startExtend,
                    pathItem.endExtend,
                    pathItem.mode,
                )
            case "p1":
                self.stretchPathItem = lshp.layoutPath(
                    QLineF(pathItem.sceneEndPoints[1], pathItem.sceneEndPoints[0]),
                    pathItem.layer,
                    pathItem.width,
                    pathItem.startExtend,
                    pathItem.endExtend,
                    pathItem.mode,
                )
        self.stretchPathItem.stretch = True
        self.stretchPathItem.name = pathItem.name

        addDeleteStretchNetCommand = us.addDeleteShapeUndo(
            self, self.stretchPathItem, pathItem
        )
        self.undoStack.push(addDeleteStretchNetCommand)

    def findClosestFontSize(self, sizes: List[int], target: int = 16) -> int:
        return min(sizes, key=lambda x: abs(x - target))

    def setRulerFont(self, target_size: int = 16) -> QFont:
        fontDatabase = QFontDatabase()
        fixedFamilies = None
        if fontDatabase:
            fixedFamilies = [
                family
                for family in fontDatabase.families(QFontDatabase.Latin)
                if fontDatabase.isFixedPitch(family)
            ]

        if not fixedFamilies:
            self.logger.warning("No fixed-pitch fonts found. Using default font.")
            return QFont()
        else:
            for family in fixedFamilies:
                styles = fontDatabase.styles(family)
                if not styles:
                    continue

                style = styles[0]  # Use the first available style
                sizes = fontDatabase.pointSizes(family, style)

                if sizes:
                    closest_size = self.findClosestFontSize(sizes, target_size)
                    font = QFont(family)
                    font.setStyleName(style)
                    font.setPointSize(closest_size)
                    font.setKerning(False)
                    return font

        self.logger.warning("No suitable font found. Using default font.")
        return QFont()

    def updateItem(self, item: QGraphicsItem):
        # update the item to the latest version of the layout cell if it is edited in another editor window
        if isinstance(item, lshp.layoutInstance):
            libItem = libm.getLibItem(
                self.editorWindow.libraryView.libraryModel, item.libraryName
            )
            cellItem = libm.getCellItem(libItem, item.cellName)
            viewItem = libm.getViewItem(cellItem, item.viewName)
            viewItemTuple = ddef.viewItemTuple(libItem, cellItem, viewItem)
            if viewItem:
                newItem = self.instLayout(viewItemTuple)
                newItem.instanceName = item.instanceName
                newItem.setPos(item.pos())
                newItem.setRotation(item.rotation())
                newItem.setScale(item.scale())
                newItem.setTransform(item.transform())
                updateRect = item.boundingRect()
                self.removeItem(item)
                self.addItem(newItem)
                self.update(updateRect)

    def renumberInstances(self):
        self.messageLine.setText("Renumbering instances. Cannot be undone.")
        instanceList = [
            item
            for item in self.items()
            if isinstance(item, (lshp.layoutInstance, lshp.layoutPcell))
        ]

        for index, layoutInstance in enumerate(instanceList):
            layoutInstance.counter = index
            if layoutInstance.instanceName.startswith("I"):
                # Preserve array notation like I2<0:5>
                if (
                        "<" in layoutInstance.instanceName
                        and ">" in layoutInstance.instanceName
                ):
                    array_part = layoutInstance.instanceName[
                        layoutInstance.instanceName.find("<"):
                    ]
                    layoutInstance.instanceName = f"I{index}{array_part}"
                else:
                    layoutInstance.instanceName = f"I{index}"
        self.instanceCounter = index + 1
        self.saveLayoutCell(self.editorWindow.file)
        self.reloadScene()



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting) a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#
import os
import pathlib
from pathlib import Path

import numpy as np
from PySide6.QtCore import (Signal, Qt, QModelIndex)
from PySide6.QtGui import (
    QPainter,
    QStandardItemModel,
    QStandardItem,
    QBrush,
    QColor,
    QPixmap,
    QImage,
)
from PySide6.QtWidgets import (QTableView, QStyledItemDelegate,
                               QStyle)

from revedaEditor.backend.pdkLoader import importPDKModule

fabproc = importPDKModule('process')
laylyr = importPDKModule('layoutLayers')


class layerDataModel(QStandardItemModel):
    _file_content_cache = {}
    _pixmap_cache = {}

    def __init__(self, data: list):
        super().__init__()
        self._data = data or []
        self.setColumnCount(5)  # Set the number of columns

        # Set the headers for the columns
        self.setHeaderData(0, Qt.Orientation.Horizontal, "")
        self.setHeaderData(1, Qt.Orientation.Horizontal, "Layer")
        self.setHeaderData(2, Qt.Orientation.Horizontal, "Purp.")
        self.setHeaderData(3, Qt.Orientation.Horizontal, "V")
        self.setHeaderData(4, Qt.Orientation.Horizontal, "S")

        for row, layer in enumerate(self._data):
            self.insertRow(row)
            # bitmap = QBitmap.fromImage(QPixmap(layer.btexture).scaled(5, 5).toImage())
            reveda_pdk_path = os.environ.get("REVEDA_PDK_PATH", None)
            if reveda_pdk_path is None:
                reveda_pdk_pathobj = Path(__file__).parents[2].joinpath(
                    "defaultPDK")
            else:
                reveda_pdk_pathobj = pathlib.Path(reveda_pdk_path)

            texturePath = reveda_pdk_pathobj.joinpath(layer.btexture)
            _pixmap = QPixmap.fromImage(self.createImage(texturePath,
                                                         layer.bcolor, 1))
            # Create a brush with black background
            brush = QBrush(QColor('black'))
            # Set the texture pattern over the black background
            brush.setTexture(_pixmap)
            item = QStandardItem()
            item.setBackground(brush)
            self.setItem(row, 0, item)
            self.setItem(row, 1, QStandardItem(layer.name))
            self.setItem(row, 2, QStandardItem(layer.purpose))
            item = QStandardItem()
            item.setCheckable(True)
            item.setCheckState(Qt.Checked if layer.selectable else Qt.Unchecked)
            self.setItem(row, 3, item)
            item = QStandardItem()
            item.setCheckable(True)
            item.setCheckState(Qt.Checked if layer.visible else Qt.Unchecked)
            self.setItem(row, 4, item)

    def createData(self, layerlist: list) -> list:
        [
            self._data.append(
                (
                    layer.name,
                    layer.visible,
                    layer.selectable,
                    layer.btexture,
                    layer.bcolor,
                )
            )
            for layer in layerlist
        ]

    @classmethod
    def readFileContent(cls, filePath):
        if filePath not in cls._file_content_cache:
            try:
                with open(filePath, "r") as file:
                    cls._file_content_cache[filePath] = file.read()
            except FileNotFoundError:
                print(f"Error: Stipple not found: {filePath}")
                return ""
            except Exception as e:
                print(f"Error reading Stipple file {filePath}: {e}")
                return ""
        return cls._file_content_cache[filePath]

    @classmethod
    def createImage(cls, filePath: Path, color: QColor, scale: int = 1):
        content = cls.readFileContent(str(filePath))

        # Use numpy's loadtxt for faster parsing of text data
        data = np.loadtxt(content.splitlines(), dtype=np.uint8)

        # Scale up the pattern by repeating each pixel
        data_scaled = np.repeat(np.repeat(data, scale, axis=0), scale, axis=1)

        height, width = data_scaled.shape

        # Create QImage with Format_ARGB32 (not premultiplied)
        image = QImage(width, height, QImage.Format.Format_ARGB32)
        # Fill with transparent pixels first
        image.fill(Qt.black)

        # Create painter to draw on the image
        painter = QPainter(image)
        painter.setPen(Qt.NoPen)
        painter.setBrush(QBrush(color))

        # Draw solid rectangles for each pixel that should be colored
        for i in range(height):
            for j in range(width):
                if data_scaled[i, j] == 1:  # Draw colored pixel
                    painter.drawRect(j, i, 1, 1)

        painter.end()
        return image


class TextureDelegate(QStyledItemDelegate):
    def paint(self, painter, option, index):
        if index.column() == 0 and option.state & QStyle.State_Selected:
            # Get the original brush from the model
            brush = index.data(Qt.BackgroundRole)
            if brush:
                # Create a semi-transparent version
                painter.fillRect(option.rect, brush)
                # Add selection overlay with transparency
                selectionColor = option.palette.highlight().color()
                selectionColor.setAlpha(100)  # Adjust alpha as needed
                painter.fillRect(option.rect, selectionColor)
            return
        super().paint(painter, option, index)


class layerViewTable(QTableView):
    columnTexture = 0
    columnName = 1
    columnPurpose = 2
    columnVisible = 3
    columnSelectable = 4

    dataSelected = Signal(str, str)
    layerSelectable = Signal(str, str, bool)
    layerVisible = Signal(str, str, bool)

    def __init__(self, parent=None, model: layerDataModel = None):
        super().__init__(parent)
        self._model = model
        self.parent = parent
        self.layoutScene = self.parent.scene
        self.setModel(self._model)

        self.setupUi()
        self.connectSignals()

    def setupUi(self):
        """Initialize UI components"""
        self.selectedRow: int = -1
        self.resizeColumnsToContents()
        self.setShowGrid(False)
        self.setSelectionBehavior(QTableView.SelectRows)
        self.setSelectionMode(QTableView.SingleSelection)
        self.verticalHeader().setVisible(False)
        # Set custom delegate for texture column
        self.setItemDelegateForColumn(self.columnTexture, TextureDelegate(self))

    def connectSignals(self):
        """Connect signal handlers"""
        self.selectionModel().selectionChanged.connect(self.onSelectionChanged)
        self._model.dataChanged.connect(self.onDataChanged)

    def getLayerInfo(self, row: int) -> tuple[str, str]:
        """Helper method to get layer name and purpose"""
        return (
            self._model.item(row, self.columnName).text(),
            self._model.item(row, self.columnPurpose).text()
        )

    def onDataChanged(self, topLeft: QModelIndex, bottomRight: QModelIndex, roles: list):
        if Qt.CheckStateRole not in roles:
            return

        row, column = topLeft.row(), topLeft.column()
        item = self._model.item(row, column)
        isChecked = item.checkState() == Qt.Checked
        layerName, layerPurpose = self.getLayerInfo(row)

        if column == self.columnSelectable:
            self.layerSelectable.emit(layerName, layerPurpose, isChecked)
        elif column == self.columnVisible:
            self.layerVisible.emit(layerName, layerPurpose, isChecked)

    def onSelectionChanged(self, selected, deselected):
        if selected.indexes():
            indices = selected.indexes()
            layerName = self._model.data(indices[self.columnName])
            layerPurpose = self._model.data(indices[self.columnPurpose])
            self.dataSelected.emit(layerName, layerPurpose)

    def updateAllLayers(self, visible: bool = None, selectable: bool = None):
        """Helper method to update all layers' visibility or selectability"""
        if visible is not None:
            state, column = Qt.Checked if visible else Qt.Unchecked, self.columnVisible
            for layer in laylyr.pdkAllLayers:
                layer.visible = visible
        else:
            state, column = Qt.Checked if selectable else Qt.Unchecked, self.columnSelectable
            for layer in laylyr.pdkAllLayers:
                layer.selectable = selectable
            # Update scene items selectability
            for item in self.layoutScene.items():
                if item.parentItem() is None and hasattr(item, 'layer'):
                    item.setEnabled(selectable)

        for row in range(self._model.rowCount()):
            self._model.item(row, column).setCheckState(state)

    def noLayersVisible(self):
        self.updateAllLayers(visible=False)

    def allLayersVisible(self):
        self.updateAllLayers(visible=True)

    def noLayersSelectable(self):
        self.updateAllLayers(selectable=False)

    def allLayersSelectable(self):
        self.updateAllLayers(selectable=True)



import pytest
from PySide6.QtCore import (QPoint, QLineF)
from PySide6.QtWidgets import (QGraphicsScene, QGraphicsView)

from revedaEditor.common.net import schematicNet, netNameStrengthEnum


@pytest.fixture
def setup_nets(qtbot):
    net1 = schematicNet(QPoint(0, 0), QPoint(10, 10))
    net2 = schematicNet(QPoint(10, 10), QPoint(20, 20))
    view = QGraphicsView()
    qtbot.addWidget(view)
    scene = QGraphicsScene()
    view.setScene(scene)
    scene.addItem(net1)
    scene.addItem(net2)
    return net1, net2, scene


def test_inheritNetName_set_to_noname(setup_nets):
    net1, net2, scene = setup_nets
    net1.nameStrength = netNameStrengthEnum.SET
    net1.name = "Net1"
    net2.nameStrength = netNameStrengthEnum.NONAME
    result = net1.inheritNetName(net2)
    assert result
    assert net2.name == "Net1"
    assert net2.nameStrength == netNameStrengthEnum.INHERIT


def test_inheritNetName_set_to_inherit(setup_nets):
    net1, net2, scene = setup_nets
    net1.nameStrength = netNameStrengthEnum.SET
    net1.name = "Net1"
    net2.nameStrength = netNameStrengthEnum.INHERIT
    result = net1.inheritNetName(net2)
    assert result
    assert net2.name == "Net1"
    assert net2.nameStrength == netNameStrengthEnum.INHERIT


def test_inheritNetName_set_to_set_conflict(setup_nets):
    net1, net2, scene = setup_nets
    net1.nameStrength = netNameStrengthEnum.SET
    net1.name = "Net1"
    net2.nameStrength = netNameStrengthEnum.SET
    net2.name = "Net2"
    result = net1.inheritNetName(net2)
    assert not result
    assert net1.nameConflict
    assert net2.nameConflict


def test_inheritNetName_inherit_to_noname(setup_nets):
    net1, net2, scene = setup_nets
    net1.nameStrength = netNameStrengthEnum.INHERIT
    net1.name = "Net1"
    net2.nameStrength = netNameStrengthEnum.NONAME
    net2.name = "Net2"
    result = net1.inheritNetName(net2)
    assert result
    assert net2.name == "Net1"
    assert net2.nameStrength == netNameStrengthEnum.INHERIT


def test_inheritNetName_inherit_to_weak(setup_nets):
    net1, net2, scene = setup_nets
    net1.nameStrength = netNameStrengthEnum.INHERIT
    net1.name = "Net1"
    net2.nameStrength = netNameStrengthEnum.WEAK
    net2.name = "Net2"
    result = net1.inheritNetName(net2)
    assert result
    assert net2.name == "Net1"
    assert net2.nameStrength == netNameStrengthEnum.INHERIT


def test_inheritNetName_inherit_to_inherit(setup_nets):
    net1, net2, scene = setup_nets
    net1.nameStrength = netNameStrengthEnum.INHERIT
    net1.name = "Net1"
    net2.nameStrength = netNameStrengthEnum.INHERIT
    net2.name = "Net2"
    result = net1.inheritNetName(net2)
    assert not result
    assert net1.nameConflict
    assert net2.nameConflict


def test_inheritNetName_inherit_to_set(setup_nets):
    net1, net2, scene = setup_nets
    net1.nameStrength = netNameStrengthEnum.INHERIT
    net1.name = "Net1"
    net2.nameStrength = netNameStrengthEnum.SET
    net2.name = "Net2"
    result = net1.inheritNetName(net2)
    assert result
    assert net1.name == "Net2"
    assert net1.nameStrength == netNameStrengthEnum.INHERIT


def test_inheritNetName_weak_to_noname(setup_nets):
    net1, net2, scene = setup_nets
    net1.nameStrength = netNameStrengthEnum.WEAK
    net1.name = "Net1"
    net2.nameStrength = netNameStrengthEnum.NONAME
    result = net1.inheritNetName(net2)
    assert result
    assert net2.name == "Net1"
    assert net2.nameStrength == netNameStrengthEnum.WEAK


def test_inheritNetName_weak_to_weak_different_name(setup_nets):
    net1, net2, scene = setup_nets
    net1.nameStrength = netNameStrengthEnum.WEAK
    net1.name = "Net1"
    net2.nameStrength = netNameStrengthEnum.WEAK
    net2.name = "Net2"
    result = net1.inheritNetName(net2)
    assert not result


def test_inheritNetName_weak_to_same_name(setup_nets):
    net1, net2, scene = setup_nets
    net1.nameStrength = netNameStrengthEnum.WEAK
    net1.name = "Net1"
    net2.nameStrength = netNameStrengthEnum.WEAK
    net2.name = "Net1"
    result = net1.inheritNetName(net2)
    assert result
    assert net2.name == "Net1"
    assert net2.nameStrength == netNameStrengthEnum.WEAK


def test_inheritNetName_weak_to_inherit(setup_nets):
    net1, net2, scene = setup_nets
    net1.nameStrength = netNameStrengthEnum.WEAK
    net1.name = "Net1"
    net2.nameStrength = netNameStrengthEnum.INHERIT
    net2.name = "Net2"
    result = net1.inheritNetName(net2)
    assert result
    assert net1.name == "Net2"
    assert net1.nameStrength == netNameStrengthEnum.INHERIT


def test_inheritNetName_weak_to_set(setup_nets):
    net1, net2, scene = setup_nets
    net1.nameStrength = netNameStrengthEnum.WEAK
    net1.name = "Net1"
    net2.nameStrength = netNameStrengthEnum.SET
    net2.name = "Net2"
    result = net1.inheritNetName(net2)
    assert result
    assert net1.name == "Net2"
    assert net1.nameStrength == netNameStrengthEnum.INHERIT


def test_inheritNetName_noname_to_inherit(setup_nets):
    net1, net2, scene = setup_nets
    net1.nameStrength = netNameStrengthEnum.NONAME
    net1.name = ""
    net2.nameStrength = netNameStrengthEnum.INHERIT
    net2.name = "Net2"
    result = net1.inheritNetName(net2)
    assert result
    assert net1.name == "Net2"
    assert net1.nameStrength == netNameStrengthEnum.INHERIT


def test_inheritNetName_noname_to_set(setup_nets):
    net1, net2, scene = setup_nets
    net1.nameStrength = netNameStrengthEnum.NONAME
    net1.name = ""
    net2.nameStrength = netNameStrengthEnum.SET
    net2.name = "Net2"
    result = net1.inheritNetName(net2)
    assert result
    assert net1.name == "Net2"
    assert net1.nameStrength == netNameStrengthEnum.INHERIT


def test_inheritNetName_noname_to_noname(setup_nets):
    net1, net2, scene = setup_nets
    net1.nameStrength = netNameStrengthEnum.NONAME
    net1.name = ""
    net2.nameStrength = netNameStrengthEnum.NONAME
    net2.name = ""
    result = net1.inheritNetName(net2)
    assert result


def test_findOverlapNets_no_overlap(setup_nets):
    net1, net2, scene = setup_nets
    net1.draftLine = QLineF(QPoint(0, 0), QPoint(0, 10))
    net2.draftLine = QLineF(QPoint(0, 20), QPoint(0, 30))
    overlap_nets = net1.findOverlapNets()
    assert len(overlap_nets) == 0


def test_findOverlapNets_with_overlap(setup_nets):
    net1, net2, scene = setup_nets
    net1.draftLine = QLineF(QPoint(0, 0), QPoint(20, 0))
    net2.draftLine = QLineF(QPoint(10, 0), QPoint(60, 0))
    overlap_nets = net1.findOverlapNets()
    assert len(overlap_nets) == 1
    assert net2 in overlap_nets


def test_findOverlapNets_multiple_overlaps(setup_nets):
    net1, net2, scene = setup_nets
    net3 = schematicNet(QPoint(5, 5), QPoint(15, 15))
    scene.addItem(net3)
    net1.draftLine = QLineF(QPoint(0, 0), QPoint(0, 20))
    net2.draftLine = QLineF(QPoint(0, 10), QPoint(0, 30))
    net3.draftLine = QLineF(QPoint(-10, 10), QPoint(10, 10))
    overlap_nets = net1.findOverlapNets()
    assert len(overlap_nets) == 2
    assert net2 in overlap_nets
    assert net3 in overlap_nets


if __name__ == "__main__":
    pytest.main()



from unittest.mock import Mock

import pytest

from revedaEditor.gui.schematicEditor import schematicEditor, \
    xyceNetlist  # adjust import path as needed


class TestRecursiveNetlisting:
    @pytest.fixture
    def mock_schematic(qtbot):
        schematic = schematicEditor()
        qtbot.addWidget(schematic)
        return schematic

    @pytest.fixture
    def mock_cir_file(self):
        return Mock()

    def test_recursive_netlisting_basic_flow(self, mock_schematic, mock_cir_file):
        # Arrange
        netlister = xyceNetlist()  # replace with your actual class
        mock_scene = mock_schematic.centralW.scene

        # Setup return values
        mock_scene.findSceneSymbolSet.return_value = [Mock(), Mock()]  # two dummy symbols

        # Act
        netlister.recursiveNetlisting(mock_schematic, mock_cir_file)

        # Assert
        # Verify all required methods were called in correct order
        mock_scene.nameSceneNets.assert_called_once()
        mock_scene.findSceneSymbolSet.assert_called_once()
        mock_scene.generatePinNetMap.assert_called_once()

        # Verify processElementSymbol was called for each symbol
        assert mock_scene.findSceneSymbolSet.return_value
        assert netlister.processElementSymbol.call_count == 2

    def test_recursive_netlisting_empty_schematic(self, mock_schematic, mock_cir_file):
        # Arrange
        netlister = xyceNetlist()
        mock_scene = mock_schematic.centralW.scene
        mock_scene.findSceneSymbolSet.return_value = []

        # Act
        netlister.recursiveNetlisting(mock_schematic, mock_cir_file)

        # Assert
        mock_scene.nameSceneNets.assert_called_once()
        mock_scene.findSceneSymbolSet.assert_called_once()
        mock_scene.generatePinNetMap.assert_called_once()
        assert not netlister.processElementSymbol.called

    def test_recursive_netlisting_error_handling(self, mock_schematic, mock_cir_file):
        # Arrange
        netlister = xyceNetlist()
        mock_scene = mock_schematic.centralW.scene
        mock_scene.nameSceneNets.side_effect = Exception("Test error")

        # Act & Assert
        with pytest.raises(Exception) as exc_info:
            netlister.recursiveNetlisting(mock_schematic, mock_cir_file)
        assert str(exc_info.value) == "Test error"

    @pytest.mark.parametrize("symbol_count", [1, 3, 5])
    def test_recursive_netlisting_multiple_symbols(self, mock_schematic, mock_cir_file,
                                                   symbol_count):
        # Arrange
        netlister = xyceNetlist()
        mock_scene = mock_schematic.centralW.scene
        mock_symbols = [Mock() for _ in range(symbol_count)]
        mock_scene.findSceneSymbolSet.return_value = mock_symbols

        # Act
        netlister.recursiveNetlisting(mock_schematic, mock_cir_file)

        # Assert
        assert netlister.processElementSymbol.call_count == symbol_count
        for symbol in mock_symbols:
            netlister.processElementSymbol.assert_any_call(symbol, mock_schematic,
                                                           mock_cir_file)
